(* begin hide *)
Set Warnings "-notation-overridden".
From Equations Require Import Equations.
From EquationsReloaded Require Import Fin.
Set Equations Transparent.
Require Import Utf8 Lia List Omega.
From EquationsReloaded Require Import nestedrec.
Local Open Scope term.
Hint Extern 100 => progress (cbv zeta in *) : typeclass_instances.
(* end hide *)

(** ** Sized-based measure library

  We use a type class [Sized] for sizes on arbitrary types. *)

Class Sized (A : Type) := size : A -> nat.

(** For lists, we must be careful to define the sizing function so that
    it takes [SizeA] as a parameter so that later size functions can
    use it in a nested manner and satisfy the guardness check.  To do so
    we define it in a section. *)

Section list_size. Context {A : Type} {SizeA : Sized A}.

  Equations list_size : Sized (list A) by struct :=
  list_size nil := 0;
  list_size (cons x xs) := S (size x + list_size xs).
End list_size.

(* begin hide *)
Existing Instance list_size.
(* end hide *)

(** The following section derives a generic [map_size] function for [Sized] types. *)

Section map_size. Context {A : Type} {SizeA : Sized A}.

  Equations? map_size {B} (l : list A) (g : forall (x : A), size x < size l -> B) : list B :=
  map_size nil _ := nil;
  map_size (cons x xs) g := cons (g x _) (map_size xs (fun x H => g x _)).
  Proof. all:cbn;lia. Qed.

  (** It has the expected spec when the function does not use the
      size information. *)

  Lemma map_size_spec {B} (g : A -> B) (l : list A) : map_size l (fun x _ => g x) = List.map g l.
  (* begin hide *)
  Proof.
    funelim (map_size l (λ (x : A) (_ : size x < size l), g x)); simpl; trivial.
    now rewrite H.
  Qed.
  (* end hide *)

  (** This proves a stronger specification for map witnessing that
      it passes smaller arguments to its function argument [g]. *)

  Lemma map_size_transform {B} (g : A -> B) (l : list A) (P : A -> B -> Prop) :
    (forall a (Ha : size a < size l), P a (g a)) -> Forall2 P l (map g l).
  (* begin hide *)
  Proof.
    funelim (map_size l (fun x (_ : size x < size l) => g x)); simpl; trivial.
    cbn. intros. constructor; auto. apply H0. lia. apply H.
    intros. apply H0. unfold size in *; lia.
  Qed.
  (* end hide *)
End map_size.
