(* begin hide *)

From Equations Require Import Equations.
From Coq Require Import List Program.Syntax Arith Lia.
(* end hide *)
(** A common idiom of functional programming is the worker/wrapper
    pattern. It usually involves a recursive function that computes the
    result, wrapped in a toplevel function calling it with specific
    parameters. The paradigmatic example is probably list reversal,
    whose tail-recursive version can be written using a recursive local
    where clause: *)

Equations rev_acc {A} (l : list A) : list A :=
  rev_acc l := go l []
   where go : list A -> list A -> list A :=
         go [] acc := acc;
         go (hd :: tl) acc := go tl (hd :: acc).

(** A typical issue with such accumulating functions is that one has to
    write lemmas in two versions to prove properties about them, once
    about the internal [go] function and then on its wrapper.
    Using the functional elimination principle associated to [rev_acc],
    we can show both properties simultaneously. *)

Lemma rev_acc_eq : forall {A} (l : list A), rev_acc l = rev l.
Proof.
  (** We apply functional elimination on the [rev_acc l] call.
      The eliminator expects two predicates:
      one specifying the wrapper and another for the worker.
      For the wrapper, we give the expected final goal but for the worker
      we have to invent a kind of loop invariant: here that the result of
      the whole [go acc l] call is equal to [rev l ++ acc]. *)

  apply (rev_acc_elim (fun A l revaccl => revaccl = rev l)
                      (fun A _ l acc go_res => go_res = rev l ++ acc)).

  (** Functional elimination provides us with the worker property for the
      initial [go [] l] call, i.e. that it is equal to [rev l ++ []], which
      trivially gives us the result. For the worker proof itself, the
      result follows from associativity of [app] and the induction hypothesis. *)
(* begin hide *)
  all:trivial; intros *.
  + now rewrite app_nil_r.
  + now rewrite <- app_assoc.
(* end hide *)
Qed.
(* begin hide *)
(** ** The worker/wrapper and well-founded recursion

  Sometimes the natural expression of an algorithm in the worker/wrapper
  pattern requires well-founded recursion: here we take an example
  algorithm translated from Haskell whose termination is justified by a measure.
  Note that the [worker] subprogram's termination measure and
  implementation depends on the enclosing [k] argument which is captured
  in the where clause. Termination is justified by a simple arithmetic argument. *)

(* begin hide *)
Obligation Tactic := idtac.
Equations? isPrime (n : nat) : bool :=
  isPrime 0 := false; isPrime 1 := false;
  isPrime 2 := true;  isPrime 3 := true;
  isPrime k := worker 2
    where worker (n' : nat) : bool by wf (k - n') lt :=
    worker n' with ge_dec n' k :=
      { | left H := true;
        | right H := if Nat.eqb (Nat.modulo k n') 0 then false else worker (S n') }.
Proof. lia. (* Linear arithmetic reasoning *) Defined.
(* end hide *)