(** printing by %\coqdockw{by}% *)
(** printing rec %\coqdockw{rec}% *)
(* begin hide *)
From Equations Require Import Equations.
From EquationsReloaded Require Import Fin.
Require Import Omega Utf8 Arith Compare_dec.
Set Equations Transparent.
(* end hide *)

(** %\Equations% supports mutual recursion on mutual inductive types at
    the top-level, using a syntax close to vanilla %\Coq%. We demonstrate
    it on an example formalization of a term-language for λ-calculus
    with non-empty application spines and well-scoped variables.  We
    define this datatype as a mutual indexed inductive type of [term]s
    and [spine]s. Technically, we use a non-recursively uniform
    parameter here instead of a proper index for the number of free
    variables. That is, the parameter varies in the arguments of the
    constructor, e.g. for [Lam] where we introduce a fresh free variable,
    but not in their conclusions, which we even omit here for [term].
    Indices and non-uniform parameters are treated the same by %\Equations%. *)

Inductive term (n : nat) : Set := Var (f : fin n) | Lam (t : term (S n)) | App (t : term n) (l : spine n)
with spine (n : nat) : Set := tip : term n -> spine n | snoc : spine n -> term n -> spine n.

(* begin hide *)
Arguments Var {n} f.
Arguments Lam {n} t.
Arguments App {n} t l.
Arguments tip {n} t.
Arguments snoc {n} ts t.
(* end hide *)

(** Suppose we want to define capture-avoiding substitution for this
    language. We first need to define lifting of a well-scoped term with
    [n] variables into a well-scoped term with [n+1] free variables,
    shifting variables above or equal to [k] by [1].
    We handle separately the case of variables: *)

Equations lift_fin {n} (k : nat) (f : fin n) : fin (S n) :=
  lift_fin 0 f := fs f;
  lift_fin (S k) fz := fz;
  lift_fin (S k) (fs f) := fs (lift_fin k f).

(** Lifting a [fin] changes the index to make space for the new free variable.
    Using a toplevel [where] clause, one can define the
    mutually recursive lifting function on terms and spines. The
    definition is accepted by the guard checker easily. *)

Equations lift {n} (k : nat) (u : term n) : term (S n) := {
  lift k (Var f) => Var (lift_fin k f);
  lift k (Lam t) => Lam (lift (S k) t);
  lift k (App f ts) => App (lift k f) (lift_spine k ts) }
where lift_spine {n : nat} (k : nat) (t : spine n) : spine (S n) := {
  lift_spine k (tip t) => tip (lift k t);
  lift_spine k (snoc ts t) => snoc (lift_spine k ts) (lift k t) }.

(** This representation lends itself naturally to a definition of
    parallel substitution. To define it at the case of abstraction, one
    must explain how to lift a substitution of [k] variables to a
    substitution of [k+1] variables by preserving the [0th] variable and
    lifting the result of the substitution otherwise.  This
    pattern-matching involves the no-confusion principle on the natural
    number index (constructors are distinct and injective),
    but does not require UIP. *)

Equations extend_var {k l : nat} (u : fin k -> term l) (f : fin (S k)) : term (S l) :=
  extend_var u fz => Var fz ;
  extend_var u (fs f) => lift 0 (u f).

(** Again, using a toplevel [where] clause, we can define this parallel substitution,
    structurally on the term and spine. *)

Equations subst_term {k l : nat} (u : fin k -> term l) (t : term k) : term l := {
  subst_term u (Var v) => u v;
  subst_term u (Lam t) => Lam (subst_term (extend_var u) t);
  subst_term u (App t l) => App (subst_term u t) (subst_spine u l) }
where subst_spine {k l : nat} (u : fin k -> term l) (t : spine k) : spine l := {
  subst_spine u (tip t) => tip (subst_term u t);
  subst_spine u (snoc ts t) => snoc (subst_spine u ts) (subst_term u t) }.

(** %\paragraph{Testing}% We can run this program inside %\Coq% to test it.
    First we need to represent a single substitution of a variable by a closed term,
    i.e. a function of type [fin 1 -> term 0], and a substitution into terms with
    1 free variable [subst1]. This is definable by pattern-matching on the [fin _] arguments. *)

Equations subst0 (t : term 0) (f : fin 1) : term 0 := subst0 t fz => t.
Equations subst1 (t : term 1) (f : fin 2) : term 1 :=
  subst1 t fz => t;
  subst1 t (fs f) => Var fz.

(** The definition of [subst0] has a single case for [fz] as the [fs _ f'] case is
    ruled out automatically by %\Equations% which can infer that [f' :
    fin 0] is uninhabited using a single dependent case analysis (this heuristic
    could be parameterized). We can now check that substitution is indeed capture-avoiding. *)

Definition id {n} : term n := Lam (Var fz).
Definition idfree : term 1 := Lam (Var (fs fz)).
Definition Ω_body : term 1 := App (Var fz) (tip (Var fz)).
Definition Ω := Lam Ω_body.

(** We define the identity function [λ x. x], which can be lifted
    transparently to any number of free variables, [λx. y] as a term
    with one free variable, along with [Ω_body], self-application of a
    variable to itself, and the corresponding term [λ x. x x] and test: *)

Check eq_refl : subst_term (subst0 Ω) id = id. (** [--- (λx. x)[Ω/0] = λx. x] *)

Check eq_refl : subst_term (subst0 Ω) Ω_body = App Ω (tip Ω). (** [--- (x x)[Ω/0] = Ω Ω] *)


(**
  Finally, we check that substituting into a non-closed term does not capture bound variables: *)

Check eq_refl : subst_term (subst1 (Var fz)) (lift 1 idfree) = idfree. (** [--- (λ x. y)[z/y] = λx. z] *)
(* begin hide *)
Check eq_refl : subst_term (subst1 id) (lift 1 idfree) = Lam id. (** [--- (λ x. y)[id/y] = λx. id] *)
(* end hide *)
