(* begin hide *)
Set Warnings "-notation-overridden".
From Equations Require Import Equations.
From EquationsReloaded Require Import Fin.
From Coq Require Import List Program.Syntax Arith Lia FunctionalExtensionality.
Require Import List Wellfounded.
Set Equations Transparent.
Set Universe Polymorphism.
Import Sigma_Notations.
From EquationsReloaded Require Import Sized nestedrec sized_rec.
(* end hide *)
(** ** Mutual well-founded recursion through GADTs
    %\label{sec:mutwfrec}%

  We present an instance of this pattern for the term substitution
  functions which is actually nested, but it can apply to
  arbitrary nested or mutual definitions. The [subst_ty] family
  encodes the types of our functions. *)

Inductive subst_ty : forall (A : Type) (P : A -> Type), Type :=
| tysubst : subst_ty (Σ {k l} (σ : fin k -> term l), term k) (fun a => term a.2.1)
| tysubsts : subst_ty (Σ {k l} (σ : fin k -> term l), list (term k)) (fun a => list (term a.2.1)).
(* begin hide *)
Derive Signature for subst_ty.
(* end hide *)

(** We will use a simple measure, assuming an overloaded [size] function
    and definition of [size] for terms and lists of sized objects. *)

Equations measure {A P} (t : subst_ty A P) (x : A) : nat :=
  measure tysubst (_, _, _, t) => size t;
  measure tysubsts (_, _, _, l) => size l.

(** We define the function by recursion on the abstract packed argument according to this measure.
    Using dependent pattern matching, the clauses for [tysubst] refine the argument
    and return type to match the type of [subst_term] and similarly for [tysubsts],
    we can hence do pattern-matching as usual on the actual arguments.
    Termination is easily proven by reasoning on sizes. *)

Equations? subst {A P} (t : subst_ty A P) (x : A) : P x by wf (measure t x) lt := {
  (Var v) [σ] => σ v;
  (Lam t) [σ] => Lam (t [extend_var σ]);
  (App t l) [σ] => App (t [σ]) (subst tysubsts (_, _, σ, l));
  subst t1 (k, l, σ, ts) => map_size ts (fun t Ints => t [σ] : term l) }
where "t [ σ ]" := (subst tysubst (_, _, σ, t)) : term.
Proof. all:repeat (unfold size; simp term_size); lia. Defined.
(* begin hide *)
(* FIXME bug in unfolding lemma automated proof with recursive call under binders. *)
Next Obligation.
  unfold subst. unfold_recursor_ext.
  destruct t. simpl. destruct x as [k [l [f' t']]]. destruct t'; try reflexivity.
  destruct x as [k [l [f' t']]]. simpl. f_equal.
Defined.
(* end hide *)
(* begin hide *)
