(** printing now %\coqdoctac{now}% *)
(** printing elimination %\coqdoctac{elimination}% *)
(** printing Hint %\coqdockw{Hint}% *)
(** printing Rewrite %\coqdockw{Rewrite}% *)
(** printing <= %\rightarrow% #⇐# *)
(**
   Mutual recursion can be seen as a special case of _nested_
   recursion, where an inductive type is defined mutually with a
   previously defined inductive type taking it as a parameter. %\Coq%
   natively supports the definition of nested inductive types, however
   there is little high-level support for working with such
   definitions: either when writing programs or when reasoning on
   these inductive types, the user is faced with the delicate
   representation of nested fixpoints, and the system does not derive
   expressive enough eliminators automatically.

  %\subsubsection{Structural recursion on nested types}%
  A common use-case for these types is nesting the type of lists in
  the definition of a new inductive type. Here we take the example of
  a well-scoped λ-term structure with an application constructor taking lists of
  terms as arguments (see, e.g. %\cite{DBLP:journals/pacmpl/PoulsenRTKV18}%
  for an application of well-scoped terms). *)
(* begin hide *)

Set Warnings "-notation-overridden".
From Equations Require Import Equations.
From EquationsReloaded Require Import Fin.
Require Import Omega Utf8 Arith Compare_dec List Lia.
Require Import Relation_Operators.
Arguments clos_trans [A].
Import Sigma_Notations.
Set Equations Transparent.
Import Inaccessible_Notations.
Equations lift_fin {n} (k : nat) (f : fin n) : fin (S n) :=
  lift_fin 0 f := fs f;
  lift_fin (S k) fz := fz;
  lift_fin (S k) (fs f) := fs (lift_fin k f).
Open Scope list_scope.
Derive Signature for Forall2.
Hint Constructors Forall2 : core.
Local Open Scope program_scope.
Local Open Scope equations_scope.
Arguments map {A B}.
(* end hide *)

Inductive term : nat -> Set :=
| Var {n} (f : fin n) : term n
| Lam {n} (t : term (S n)) : term n
| App {n} (t : term n) (l : list (term n)) : term n.

(* begin hide *)
Arguments Var {n} f.
Arguments Lam {n} t.
Arguments App {n} t l.
(* Declare Scope term.*)
Delimit Scope term with term.
Bind Scope term with term.
Local Open Scope term.
Derive Signature NoConfusion NoConfusionHom for term.

Equations lift {n} (k : nat) (σ : term n) : term (S n) := {
  lift k (Var f) => Var (lift_fin k f);
  lift k (Lam t) => Lam (lift (S k) t);
  lift k (App f ts) => App (lift k f) (lift_terms k ts) }
where lift_terms {n : nat} (k : nat) (t : list (term n)) : list (term (S n)) := {
  lift_terms k nil => nil;
  lift_terms k (cons t ts) => cons (lift k t) (lift_terms k ts) }.
(* end hide *)

(** Suppose we want to define capture-avoiding substitution for this
    language. We first need to define lifting of a well-scoped term with
    [n] variables into a well-scoped term with [n+1] free variables,
    shifting variables above or equal to [k] by [1]. We assume
    [lift : forall {n} (k : nat) (σ : term n) : term (S n)]
    and concentrate on substitution which is defined similarly.
    Using [lift] we can define a substitution extension function [extend_var]
    which lifts a substitution of [k] variables into a substitution of [k + 1]
    variables keeping the first variable untouched. *)

Equations extend_var {k l : nat} (σ : fin k -> term l) (f : fin (S k)) : term (S l) :=
  extend_var σ fz => Var fz ;
  extend_var σ (fs f) => lift 0 (σ f).

(** For definitions of fixpoints on nested mutual inductive types,
  %\Equations% allows users to factorize the nested fixpoint definitions in
  toplevel [where] clauses, so that one does not need to write an internal
  fixpoint construction inside the program. Multiple calls to the nested
  function can also refer to the same function, e.g if we extend our term
  structure with other constructors using lists of terms. We want to use
  the notation [t [ σ ]] for subtitution [σ] applied to [t]. We first have to
  declare it to the parser, and then bind it to its expansion using a where
  clause. *)

Reserved Notation "t [ σ ]" (at level 10).
Equations tsubst {k l : nat} (σ : fin k -> term l) (t : term k) : term l := {
  (Var v) [σ] => σ v;
  (Lam t) [σ] => Lam (t [extend_var σ]);
  (App t l) [σ] => App (t [σ]) (tsubsts σ l) }
where tsubsts {k l} (σ : fin k -> term l) (t : list (term k)) : list (term l) := {
  tsubsts σ nil => nil;
  tsubsts σ (cons t ts) => cons (t [σ]) (tsubsts σ ts) }
where "t [ σ ]" := (tsubst σ t) : term.

(** The %\Coq% kernel will check a single fixpoint definition for [tsubst] where
    [tsubsts] has been expanded at its call sites, as definitions on nested recursive
    types correspond to nested local fixpoints in CIC. *)

(** %\subsubsection{Reasoning}%

   Remark that our definition of [tsubsts] is equivalent to a call
   to [map] on lists.  %\Equations% currently needs the "expanded"
   version to properly recognize recursive calls, but one can readily
   add this equation to the [tsubst] rewrite database gathering
   the definining equations of [tsubst] to abstract away from this detail: *)

Lemma tsubsts_map k l σ t : @tsubsts k l σ t = List.map (tsubst σ) t.
(* begin hide *)
Proof. induction t; now simpl; rewrite ?IHt. Qed.
(* end hide *)
Hint Rewrite tsubsts_map : tsubst.

(** The elimination principle generated from this definition is giving a conjunction
    of two predicates as a result, and has the proper induction hypotheses for nested
    recursive calls. Given that the [tsubsts] function is essentially mapping the
    substitution, we can derive a specialized induction principle giving us
    [Forall2 P l (map (tsubst σ) l)] hypotheses for the recursive call to
    [tsubsts]. [Forall2 P l l'] is equivalent to the pointwise conjunction of [P]
    for the elements of [l] and [l']. We can derive: *)

Lemma tsubst_elim_all (P : ∀ k l : nat, (fin k → term l) → term k → term l → Prop) :
  (∀ k l σ (f : fin k), P k l σ (Var f) (σ f)) →
  (∀ k l σ (t : term (S k)), P (S k) (S l) (extend_var σ) t (t [extend_var σ]) →
    P k l σ (Lam t) (Lam (t [extend_var σ]))) →
  (∀ k l σ (t : term k) (ts : list (term k)), P k l σ t (t [σ]) →
    Forall2 (P k l σ) ts (map (tsubst σ) ts) →
    P k l σ (App t ts) (App (t [σ]) (map (tsubst σ) ts))) →
  ∀ k l σ t, P k l σ t (t [σ]).
(* begin hide *)
Proof. intros ????. apply (tsubst_elim _ (fun k l σ => Forall2 (P k l σ)));
                      intros *; simp tsubst; auto. Qed.
(* end hide *)

(** This is good, however the program still relies on the syntactic guardedness check.
    It only takes a bit of type information to get this useful reasoning principle
    directly, using well-founded recursion.

  %\subsection{Well-founded recursion on nested types}%
  Well-founded recursion requires us to give an explicit relation
  explaining why going through the list in the application case is ok.
  We will do so by relying on the fact that [map f l] can only apply [f]
  to members of [l]. Note that this is not a consequence of the
  parametricity of [map], but can easly be seen from its definition.
  To reflect this in the logic, we must define a variant of map that carries
  proofs of membership in [l] of each element passed to [f].
  Membership is a standard notion of the theory of lists, which can be defined
  inductively as follows%\footnote{We skip implicit arguments, notations,
  hints and type-class instance declarations in the following, unless they are crucial.}%: *)
(* begin hide *)
Reserved Notation " x ∈ s " (at level 70, s at level 10).
(* end hide *)

Inductive In {A} (x : A) : list A -> Prop :=
| here {xs} : x ∈ (x :: xs)
| there {y xs} : x ∈ xs -> x ∈ (y :: xs) where "x ∈ s" := (In x s).
(* begin hide *)
Arguments here {A x xs}.
Arguments there {A x y xs} _.
Derive Signature for In.
(* end hide *)

Equations mapIn {A B : Type} (l : list A) (f : forall a, a ∈ l -> B) : list B :=
  mapIn nil _ := nil;
  mapIn (cons x xs) f := cons (f x here) (mapIn xs (fun x H => f x (there H))).

(** [mapIn] is a dependently-typed variant of [map] which passes proofs of membership
    to [f]. Note that [f]'s type is refined to [forall a, a ∈ (x :: xs) -> B] in the
    second clause. In case the function does not use its argument, it
    behaves like a regular [map]. *)

Lemma mapIn_irrel {A B} (f : A -> B) l : mapIn l (fun (x : A) (_ : In x l) => f x) = List.map f l.
(* begin hide *)
Proof.
  funelim (mapIn l _); rewrite ?H; trivial.
Qed.
Hint Rewrite @mapIn_irrel : map.
(* end hide *)

(** More interestingly, [mapIn] transforms a predicate valid on all members
    of a list into a property of the list and its mapping, which is easily proven
    by functional elimination. *)

Lemma mapIn_spec {A B} (l : list A) (g : forall x : A, In x l -> B) (P : A -> B -> Prop) :
  (forall a (ina : In a l), P a (g a ina)) <-> Forall2 P l (mapIn l g).
(* begin hide *)
Proof.
  revert P.
  funelim (mapIn _ _); simpl; auto. split; auto.
  intros H; depelim H. intros a ina. depelim ina.
  split. intros. constructor; auto. apply H. intros; apply H0.
  intros H0; dependent elimination H0. intros a ina. depelim ina; auto.
  apply H. apply f0.
Qed.
(* end hide *)

(** We can also define a well-founded relation on [term] that shows that
    any member of the list in the [App] constructor is a subterm. The
    %\ind{Subterm}% derivatiom algorithm of %\Equations% does not yet recognize
    these nested types: it would produce a relation without [term_sub_3]
    here, so we define it ourselves. Note that this is an
    _heterogeneous_ relation between terms with potentially different
    numbers of free variables.  The [Var] constructor has no recursive
    subterm, while [Lam] has a direct subterm with one more free
    variable and [App] has one constructor for the substerm and one for
    the argument list. We add these constructors to a hint database that
    is used to prove recursive call obligations. *)

Inductive term_sub : ∀ {m n}, term m → term n -> Prop :=
| term_sub_1 : ∀ n (t : term (S n)), term_sub t (Lam t)
| term_sub_2 : ∀ n (t : term n) l, term_sub t (App t l)
| term_sub_3 : ∀ n (t : term n) l x, In x l -> term_sub x (App t l).
(* begin hide *)
Hint Constructors term_sub : subterm_relation.
Derive Signature for term_sub.
(* end hide *)

(** We define the actual relation that is well-founded as the transitive
    closure of [term_sub]. Note that the relation must be
    _homogeneous_ and on a closed type, so we pack the terms with their
    number of free variables in a dependent pair.  It is shown
    well-founded by a _nested_ structural recursion on the [term] and
    [list term] structure and dependent elimination of the [term_sub]
    inductive family: in the end we must always come back to %\Coq%'s
    primitive fixpoint constructions. However this must be done only
    once per datatype: this relation can be used to justify nested
    recursive definitions without ever coming back to the syntactic
    check, while still enjoying the same definitional equations
    on closed terms. For non-nested cases using our derived %\ind{Subterm}%
    relation, we even get the _same_ definitional equations, as for
    the standard %\cst{Below}% encoding. *)

Definition term_subterm := clos_trans (λ x y : (Σ n, term n), term_sub x.2 y.2).
Instance wf_term_subterm : WellFounded term_subterm.
(* begin hide *)
Proof.
  apply Transitive_Closure.wf_clos_trans.
  red. intros [a t]. revert a t. fix IH 2. destruct t;
  constructor; intros; simp_sigmas; simpl in *. depelim H.
  depelim H. apply IH.
  depelim H. apply IH.
  revert y0 l H. fix IH' 2. destruct l. clear IH IH'. intros H; depelim H.
  intros H. inversion H. apply IH.
  eapply IH'; eauto.
Defined.
(* Print Assumptions wf_term_subterm. *)
Hint Unfold term_subterm : subterm_relation.
(* end hide *)

(** We come back to our definition of substitution. This time the function
    is defined as well-founded on the subterm relation for terms, and we
    do not need to inline the definition of [mapIn]. The proofs of termination
    are solved automatically using the previously declared hints. They are trivial:
    for the nested recursive call, we must show [term_sub t (App t l)]
    under the assumption [Inl : t ∈ l]. *)

Equations tsubst2 {k l} (σ : fin k -> term l) (t : term k) : term l by wf (k, t) term_subterm := {
  (Var v) [σ] => σ v;
  (Lam t) [σ] => Lam (t [extend_var σ]);
  (App t l) [σ] => App (t [σ]) (mapIn l (fun t Inl => t [σ])) }
where "t [ σ ]" := (tsubst2 σ t) : term.

(** An eliminator equivalent to [tsubst_elim_all] is automatically derived.
    Note that [mapIn] is entirely generic for lists.
    The principle that mapping a function over a container should
    give us a proof that the elements passed to the client do belong
    to the container is essential here. In particular this idea would
    also work with an abstract container type (e.g. sets) if it provided
    a similarly strongly specified [map] function. This is proof dependent
    programming, but of course the extraction of [mapIn] does _not_ carry
    membership proofs, as [In] is a _proposition_ here (a [Prop]).

    In a system with sized types, a direct call to regular [map] would
    be allowed here, but requires the [term] datatype and substitution
    function to be indexed by sizes%\footnote{\equationsurl{material/agda/nested\_sized.agda formalizes this}}%.
    With size annotations the [l] argument of [App] has type [list (term j k)]
    while the initial term [t] has type [term i k], with size relation [j < i].
    The nested call becomes [map (tsubst j σ) l]: i.e. we pass
    a smaller size to the substitution. In contrast, we do the proof passing
    explicitely through [mapIn].
    We can also provide a size-based variant of [map] and recover
    a similar elimination principle for a substitution based on
    it %(see \S \ref{sec:appendix})%. *)
(* begin hide *)
Lemma tsubst2_elim_all :
  ∀ (P : ∀ k l : nat, (fin k → term l) → term k → term l → Prop),
  (∀ k l σ (f : fin k), P k l σ (Var f) (σ f)) →
  (∀ k l σ (t : term (S k)), P (S k) (S l) (extend_var σ) t (t [extend_var σ]) →
    P k l σ (Lam t) (Lam (t [extend_var σ]))) →
  (∀ k l σ (t : term k) (ts : list (term k)), P k l σ t (t [σ]) →
    Forall2 (P k l σ) ts (map (tsubst2 σ) ts) →
    P k l σ (App t ts) (App (t [σ]) (map (tsubst2 σ) ts))) →
  ∀ k l σ t, P k l σ t (t [σ]).
Proof. intros P ????. apply (tsubst2_elim _); intros *; auto.
       intros. generalize (proj1 (mapIn_spec _ _ _) H3); simp map tsubst2.
Qed.

(** %\paragraph{Testing}% We can run this program inside %\Coq% to test it.
    First we need to represent a single substitution of a variable by a closed term,
    i.e. a function of type [fin 1 -> term 0], and a substitution into terms with
    1 free variable [subst1]. This is definable by pattern-matching on the [fin _] arguments. *)

Equations subst0 (t : term 0) (f : fin 1) : term 0 := subst0 t fz := t.
Equations subst1 (t : term 1) (f : fin 2) : term 1 :=
  subst1 t fz := t;
  subst1 t (fs fz) := Var fz.

(** The definition of [subst0] has a single case for [fz] as the [fs _ f'] case is
    ruled out automatically by %\Equations% which can infer that [f' :
    fin 0] is uninhabited using a single dependent case analysis (this heuristic
    could be parameterized). We can now check that substitution is indeed capture-avoiding. *)

Definition id {n} : term n := Lam (Var fz).
Definition idfree : term 1 := Lam (Var (fs fz)).
Definition Ω_body : term 1 := App (Var fz) (cons (Var fz) nil).
Definition Ω := Lam Ω_body.

(** Definitional equations for the [Var] and [Lam] caees. *)
Check (fun n l (σ : fin n -> term l) (f : fin n) => eq_refl) : forall n l σ f, (Var f) [σ] = σ f.
Check (fun n l (σ : fin n -> term l) (f : term (S n)) => eq_refl) : forall n l σ (t : term (S n)),
    (Lam t) [σ] = Lam (t [extend_var σ]).
(** For App though we lose a definitional equality *)
Fail Check (fun n l (σ : fin n -> term l) (t : term n) (l : list (term n)) => eq_refl) :
  forall n l σ (t : term n) (l : list (term n)), (App t l) [σ] = App (t [σ]) (mapIn l (fun t Inl => t[σ])).

(** We define the identity function [λ x. x], which can be lifted
    transparently to any number of free variables, [λx. y] as a term
    with one free variable, along with [Ω_body], self-application of a
    variable to itself, and the corresponding term [λ x. x x] and test: *)

Check eq_refl : id [subst0 Ω] = id. (** [--- (λx. x)[Ω/0] = λx. x] *)

Check eq_refl : Ω_body [subst0 Ω] = App Ω (cons Ω nil). (** [--- (x x)[Ω/0] = Ω Ω] *)

(** Finally, we check that substituting into a non-closed term does not capture bound variables: *)

Check eq_refl : (lift 1 idfree) [subst1 (Var fz)] = idfree. (** [--- (λ x. y)[z/y] = λx. z] *)
Check eq_refl : (lift 1 idfree) [subst1 id] = Lam id. (** [--- (λ x. y)[id/y] = λx. id] *)
(* end hide *)