(* begin hide *)
Require Import Equations.Equations.
Require Import Equations.Prop.Logic Equations.Prop.NoConfusion.
Require Import Equations.Prop.DepElim.
Import Sigma_Notations.
Inductive vector (A : Type) : nat -> Type :=
| nil : vector A 0
| cons (a : A) {n : nat} (v : vector A n) : vector A (S n).
Arguments nil {A}.
Arguments cons {A} a {n} v.
Check noConfusion_inv.
Set Equations Transparent.
Derive Signature NoConfusion NoConfusionHom for vector.

Notation " 'rew' H 'in' c " := (@eq_rect _ _ _ c _ H) (at level 20).

Reserved Notation "p ..1" (at level 3, format "p  ..1").
Reserved Notation "p ..2" (at level 3, format "p  ..2").

Equations J {A : Type} {x : A} (P : forall y : A, x = y -> Type) {y : A} (p : x = y) (u : P x eq_refl) : P y p :=
J P eq_refl u := u.

Equations pr1_path {A : Type} {P : A -> Type} {u v : sigma P} (p : u = v) : u.1 = v.1 :=
  { eq_refl ..1 := eq_refl }
where "p ..1" := (pr1_path p).

Equations pr2_path {A : Type} {P : A -> Type} {u v : sigma P} (p : u = v) : p..1 # u.2 = v.2 :=
  { eq_refl ..2 := eq_refl }
where "p ..2" := (pr2_path p).

Polymorphic
Lemma J_inv@{i j} {A : Type@{i}}
      (B : A -> Type@{i}) (x : A) (p q : B x) (G : p = q -> Type@{j}) :
      (forall (e : (x, p) = (x, q)) (e' : e..1 = eq_refl),
          G (transport (fun ex => ex # p = q) e' e..2)) ->
  (forall e : p = q, G e).
Proof.
  intros H. intros e. destruct e. apply (H eq_refl eq_refl).
Defined.
(* end hide *)
(** ** NoConfusion and injectivity of inductive families
    %\label{sec:noconfhetero}%

  On vectors, the heterogeneous no-confusion relation is empty outside the diagonal,
  and otherwise returns an equality of the arguments of the constructors
  (or [True] when there are none).
  This is an example of the encode-decode method from Homotopy Type Theory:
  we are characterizing the equality in the vector family in an alternative way.
 *)

Equations NoConf_vector {A} (x y : Σ n, vector A n) : Prop :=
NoConf_vector (_, nil) (_, nil) := True ;
NoConf_vector (_, @cons a n v) (_, @cons a' n' v') :=
  (a, n, v) = (a', n', v') :> Σ (a : A) (n : nat), vector A n ;
NoConf_vector _ _ := False.

(** One can show that the type [noConf_vector x y] is equivalent to [x = y]:
    in the first direction, we pattern-match on all arguments, except the indices. *)

Equations noConf_vector {A} (x y : Σ n, vector A n) : NoConf_vector x y -> x = y :=
noConf_vector (_, nil) (_, nil) I := eq_refl;
noConf_vector (_, @cons a n v) (_, @cons a' n' v') eq_refl := eq_refl.

(** In the other direction, we split on the equality proof [x = y], which determines
    that both arguments are the same, hence the inaccessible patterns.
    Pattern-matching on the first vector, we discover that its indices are also
    determined uniquely (the [?(0)] and [?(S n)] inaccessible patterns witness that): *)

Equations noConf_vector_inv A (x y : Σ n, vector A n) : x = y -> NoConf_vector x y :=
noConf_vector_inv A (?(0), nil) ?((0, nil)) eq_refl := I;
noConf_vector_inv A (?(S n), @cons a n v) ?((S n, cons a v)) eq_refl := eq_refl.

(** The equivalence is also _strong_ in the sense that even for open
    terms headed by a constructor, passing a reflexivity proof [eq_refl]
    to the inverse [noConf_vector_inv] and composing with
    [noConf_vector] should produce a reflexivity proof.
    %\citet{DBLP:phd/basesearch/Cockx17}% showed that this is necessary
    for the computational behavior of definitions to be correct. Let's
    recall why this is so. The basic problem we want to solve is
    injectivity of constructors of indexed families. Concretely, the
    question is: under which conditions can we solve the following goal
    depending on an equality between vectors of the same length: *)

Lemma inject_vcons {A} n (a a' : A) (v v' : vector A n) (P : forall a' v', cons a v = cons a' v' -> Type)
      (prf : P a v eq_refl) : (forall (e : cons a v = cons a' v'), P a' v' e).
  (** We want to simplify the equality by applying injectivity of [cons],
     however me must do so in a proof-relevant way as the goal [P a' v' e] depends
     on the shape of the proof [e]. The solution of op. cit. is to first transform
     this homogeneous equality into an heterogeneous one, essentially by using
     the inverse of the [J] rule. This gives the following goal, where [e..i]
     is a notation for projecting the ith component of an equality of telescopes,
     and [e # t] is transport along [e] (a variant of [J]):
     [[
     forall (e : (S n, cons a v) = (S n, cons a' v')) (e' : e ..1 = eq_refl), P a' v' (e' # e ..2)
     ]]
   *)
  (* begin hide *)
Proof.
  refine (J_inv _ _ _ _ _ _).
  (* end hide *)
  (** We now have an heterogeneous equality between two vectors of size [S n] on
      which

      [noConf_vector] can be applied, as it is an equivalence. This gives an equivalent goal: *)
  (* begin hide *)
  refine (apply_noConfusion (A:=Σ n, vector A n) _ _ _).
  (* end hide *)
  (**
      [[
  forall (H : NoConf_vector (S n, cons a v) (S n, cons a' v'))
  (e' : (noConf_vector H) ..1 = eq_refl), P a' v' (e' # (noConf_vector H) ..2)
  ]]

      Now [NoConf_vector] can compute as we have constructors at the head: moreover they are both [cons],
      so it gives us a telescopic equality of the arguments. Note that the rest of the goal
      depends on the shape of [H], which
      should ultimately become a reflexivity proof for this to progress.
   *)
  (* begin hide *)
  Arguments noConfusion_inv : clear implicits.
  simpl NoConfusion. unfold NoConfusionPackage_vector.
  (* end hide *)
  (** [[
      forall (H : (a, n, v) = (a', n, v')) (e' : (noConf_vector H) ..1 = eq_refl :> S n = S n),
      P a' v' (e' # (noConf_vector (S n, cons a v) (S n, cons a' v') H) ..2)
   ]]

     Applying simplifications of telescopic equality and the [J] rule, we can unify [a] and [a']
     here to make a modicum of progress. But the first equality in the goal will still contain a
     proof of [n = n] that cannot be eliminated directly without UIP.
     The insight of %\citet{DBLP:conf/cpp/CockxD17}% is to use higher-dimensional unification at this point,
     concentrating on lower dimensional variants of [H] and [e']. This amounts to doing
     a nested unification, resulting in a final instantiation of [H] and [e'] by [eq_refl],
     which makes [noConf_vector] reduce to [eq_refl] as well (as this is a %strong% equivalence),
     allowing to close the proof using [prf].
     In other words, it really suffises to show that [P] holds for the same two vectors and
     a proof of reflexivity. *)
(* begin hide *)
Abort.
(* end hide *)
