(* begin hide *)
Require Import Equations.Equations.
Require Import Equations.Prop.Logic Equations.Prop.NoConfusion.
Require Import Equations.Prop.DepElim.
Import Sigma_Notations.
Inductive vector (A : Type) : nat -> Type :=
| nil : vector A 0
| cons (a : A) {n : nat} (v : vector A n) : vector A (S n).
Arguments nil {A}.
Arguments cons {A} a {n} v.
Set Equations Transparent.
Derive Signature NoConfusion NoConfusionHom for vector.
(* bug with innaccessible checking, weird *)

Notation " 'rew' H 'in' c " := (@eq_rect _ _ _ c _ H) (at level 20).

Reserved Notation "p ..1" (at level 3).
Reserved Notation "p ..2" (at level 3).

Equations J {A : Type} {x : A} (P : forall y : A, x = y -> Type) {y : A} (p : x = y) (u : P x eq_refl) : P y p :=
J P eq_refl u := u.

Equations pr1_path {A : Type} {P : A -> Type} {u v : sigma P} (p : u = v) : u.1 = v.1 :=
  { eq_refl ..1 := eq_refl }
where "p ..1" := (pr1_path p).

Equations pr2_path {A : Type} {P : A -> Type} {u v : sigma P} (p : u = v) : p..1 # u.2 = v.2 :=
  { eq_refl ..2 := eq_refl }
where "p ..2" := (pr2_path p).

Polymorphic
Lemma J_inv@{i j} {A : Type@{i}}
      (B : A -> Type@{i}) (x : A) (p q : B x) (G : p = q -> Type@{j}) :
      (forall (e : (x, p) = (x, q)) (e' : e..1 = eq_refl),
          G (transport (fun ex => ex # p = q) e' e..2)) ->
  (forall e : p = q, G e).
Proof.
  intros H. intros e. destruct e. apply (H eq_refl eq_refl).
Defined.
(* end hide *)

(** ** A homogeneous no-confusion principle
    %\label{sec:noconfusionhom}% *)
(** Actually, relying on higher-dimensional unification for this problem
    is overkill.  We can rather devise a more precise _homogeneous_
    no-confusion principle working on any two terms in the _same
    instance_ of the inductive family to directly solve injectivity
    goals between constructors. Our analysis of the problem is that the
    heterogeneous no-confusion principle is too general. By comparing
    constructors in different instances of the same family, it does not
    rely on the index structure of each constructor, while the problems
    of injectivity we face are always between objects in the same
    instance of the familiy, i.e. they share the exact same indices.
    This means that we can reduce the content of proofs in the no-confusion
    notion while still maintaining an equivalence with
    propositional equality.  On the [cons] diagonal case of vectors for
    example, we can just ask for equalities between the head and the
    tail of the vector, while removing the problematic equality between
    the natural number arguments.

    %\paragraph{Forced arguments}% Our homogeneous no-confusion
    principle essentially relies on the analysis of forced arguments of
    constructors, a notion pioneered by
    %\citet{DBLP:conf/types/BradyMM03}% and recently used in
    %\cite{gilbert:hal-01859964}% to provide a characterization of
    strict inductive propositions.
    In essence, an argument [x]
    of a constructor %\cstr{c}% is forced if it appears in %\cstr{c}%'s conclusion
    in a pattern position. For example, the [n] argument of [cons] appears under a successor
    in [cons]'s conclusion [vector A (S n)], so it is forced.
    Initially this information was used to analyse which arguments could
    be erased at compile-time, but here can make use of this in type theory
    and internalize this notion through a type equivalence.
    For vectors, the homogeneous no-confusion principle is defined as: *)

Equations NoConfHom_vector {A n} (x y : vector A n) : Prop :=
NoConfHom_vector nil nil := True ;
NoConfHom_vector (@cons a n v) (@cons a' n v') := (a, v) = (a', v') :> Σ (_ : A), vector A n.

(** This relies on dependent pattern-matching for the index [n] of vectors:
    by discriminating on it, %\Equations% can see that there are only two cases
    to consider: if both vectors are empty or both have the same size [n]
    (we don't need to write inaccessible patterns for the [n] variables, as proven
    by %\citet{DBLP:journals/pacmpl/CockxA18}%). In case they are both [cons]
    we just need to record the equality of the heads and tails, which actually
    degenerates into an equality between non-dependent pairs. One can show again
    that [NoConfHom_vector x y] is equivalent to [x = y] by dependent pattern-matching.
    We just show the forward direction, where we can see that [n] is inaccessible.
 *)

Equations noconf_vector {A} n (x y : vector A n) : NoConfHom_vector x y -> x = y :=
noconf_vector ?(0) nil nil I := eq_refl;
noconf_vector ?(S n) (@cons a n v) (@cons a' n v') eq_refl := eq_refl.

(* begin hide *)
Equations noconf_vector_inv {A n} (x y : vector A n) : x = y -> NoConfHom_vector x y :=
noconf_vector_inv nil ?(nil) eq_refl := I;
noconf_vector_inv (@cons a n v) ?(@cons _ a n v) eq_refl := eq_refl.

Definition NoConfHom_vector_package A n : NoConfusionPackage (vector A n).
  refine {| NoConfusion := NoConfHom_vector;
     noConfusion := noconf_vector n;
     noConfusion_inv := noconf_vector_inv |}.
Proof.
  intros a b e. depelim a; depelim b; depelim e; reflexivity.
  intros a b e. depelim a; depelim b; depelim e; reflexivity.
Defined.

Lemma noConf_vector_inv' {A n} :
  forall (a b : vector A n) (e : NoConfHom_vector a b), noconf_vector_inv a b (noconf_vector n a b e) = e.
Proof.
  destruct a; intros b; dependent elimination b;
    intros e; dependent elimination e; reflexivity.
Defined.

Definition noConf_vector_strong {A n} (a : A) (v : vector A n) :
    noconf_vector (S n) (cons a v) (cons a v) (noconf_vector_inv (cons a v) (cons a v) eq_refl) = eq_refl := eq_refl.

Definition noConf_vector_strong_inv {A n} (a : A) (v : vector A n) :
    noconf_vector_inv (cons a v) (cons a v) (noconf_vector (S n) (cons a v) (cons a v) eq_refl) = eq_refl := eq_refl.

(** Crucially, the two definitions do not use UIP themselves, so it is easy to check
    that these functions form again a _strong_ equivalence. The [inject_vcons]
    proof can be solved by applying homogeneous no confusion directly to the initial goal. *)

Lemma inject_vcons {A} n (x y : A) (v v' : vector A n) (P : forall y v', cons x v = cons y v' -> Type)
      (prf : P x v eq_refl) : (forall (e : cons x v = cons y v'), P y v' e).
Proof. refine (apply_noConfusion (A:=vector A (S n)) _ _ _); simplify *. exact prf. Qed.
