(* begin hide *)
From Equations Require Import Equations.
(* end hide *)
(** In %\citet{DBLP:conf/cpp/CockxD17}%, dependent elimination was solved
    by relying on a notion of higher-dimensional unification. We depart
    from this solution by relying instead on an homogeneous no-confusion
    principle, giving a simpler explanation of dependent
    pattern-matching compilation. No-confusion is the idea that
    constructors of inductive families are both discriminable and
    injective. Our solution is more limited in the sense that we do not
    handle higher-dimensional equations that can arise for example when
    pattern-matching on the equality type, e.g. for HoTT-style
    reasoning, but it does apply to indexed inductive families for which
    pattern-matching does not require %\ind{UIP}%. We defer a detailed
    presentation of no-confusion and our solution to section
    %\ref{sec:noconfusionhom}%, and first motivate why an elaborate
    solution is needed.

    We take an example inspired by the work on Exceptional Type Theory
    %\cite{pedrot:hal-01840643}%.  This extension of CIC introduces a
    notion of effect in the type theory, namely call-by-name
    exceptions. When working in the impure fragment of that theory, all
    inductive types have an additional [raise_ℕ] argument for a
    parametric type of exception names [E]. For example, natural numbers
    become: *)

Inductive ℕ (E : Type) : Type :=
| O : ℕ E | S : ℕ E -> ℕ E
| raise_ℕ : E -> ℕ E.
(* begin hide *)
Derive NoConfusion for ℕ.
Arguments O {_}.
Arguments S {_} _.
Arguments raise_ℕ {_} _.
(* end hide *)

(** If the exception type is empty, then this type is equivalent to the natural numbers.
    In the [raise_ℕ] case we use the _empty pattern_ [!] to indicate a variable with empty type, here [False]. *)

Equations N_empty : ℕ False -> nat :=
N_empty O := Datatypes.O;
N_empty (S n) := Datatypes.S (N_empty n);
N_empty (raise_ℕ !).

(** Exceptional type theory handles the whole of CIC, hence
    effectful values can also appear in indices of inductive families,
    which themselves contain exceptions. E.g., vectors become [Vec]: *)

Inductive Vec E (A : Type) : ℕ E -> Type :=
| nil  : Vec E A O | cons : forall {n} (x : A) (xs : Vec E A n), Vec E A (S n)
| raise_vec : forall (e : E), Vec E A (raise_ℕ e).
(* begin hide *)
Arguments nil {_ _}.
Arguments cons {_ _ _} _ _.
Arguments raise_vec {_ _} _.
Derive Signature NoConfusion NoConfusionHom for Vec.
(* end hide *)

(** In op. cit., the authors develop an exceptional parametricity
    translation that can be used to show that translating any pure term
    in the effectful calculus produces an effect-free term. To do so,
    the parametricity translation produces inductive families expressing
    validity of a given potentialy effectful term. Validity is exception-freeness
    in this case. Below is the definition of the parametricity
    relation for effectful vectors: it explictely carves out exception
    free vectors and indices. *)

Inductive Vec_param {E A} : forall (n : ℕ E), Vec E A n -> Type :=
| vnil_param : Vec_param O nil
| vcons_param : forall (n : ℕ E) (a : A) (v : Vec E A n), Vec_param n v -> Vec_param (S n) (cons a v).

(** The exceptional parametricity translation produces validity proofs
    for any pure term of CIC, but as soon as one wants to reason
    on locally effectful terms to embed them in globally pure ones,
    one must reason with the validity proofs. This is where dependent
    pattern-matching can get in the way. Typically, one would like
    to show that if a non-empty vector is parametric then its tail will
    be as well: *)
(* begin hide *)
Derive Signature NoConfusionHom for Vec_param.
(* end hide *)

Equations param_tl {E A} a n (v : Vec E A n) (X : Vec_param (S n) (cons a v)) : Vec_param n v :=
  param_tl a n v (vcons_param a n v X) := X.

(** Under the hood, dependent pattern-matching compilation needs to
    solve a constraint of the form [cons a n v = cons a' n v' :> Vec E
    A (S n)] into the simpler constraint [(a, v) = (a', v')] using
    injectivity of the [cons] constructor. After that, the arguments
    [a], [a'] and [v], [v'] can be unified to prove that the
    [vcons_param a n v X] is the only way to introduce a proof of
    [Vec_param (S n) (cons a v)]. It is this first seemingly trival
    simplification step between indexed vectors that, if done naïvely,
    would introduce UIP. Note that none of the indices here have decidable
    equality as [E] is abstract, so we cannot rely on a potential proof of UIP
    for [Vec]. *)

(* begin hide *)
Definition foo := param_tl_elim :
  forall
    P : forall (E A : Type) (a : A) (n : ℕ E) (v : Vec E A n),
      Vec_param (S n) (cons a v) -> Vec_param n v -> Prop,

    (forall (E A : Type) (a0 : A) (n0 : ℕ E) (v0 : Vec E A n0) (v1 : Vec_param n0 v0),
        P E A a0 n0 v0 (vcons_param n0 a0 v0 v1) v1) ->

    forall (E A : Type) (a : A) (n : ℕ E) (v : Vec E A n) (X : Vec_param (S n) (cons a v)),
      P E A a n v X (param_tl a n v X).

Print Assumptions param_tl.
(* end hide *)