(* begin hide *)
Set Warnings "-notation-overridden".
From Equations Require Import Equations. From EquationsReloaded Require Import Fin.
From Coq Require Import List Program.Syntax Arith Lia.
Require Import List Wellfounded.
Set Equations Transparent.
Set Universe Polymorphism.
Import Sigma_Notations.
Notation "'Σ' x .. y , P" := (sigma _ (fun x => .. (sigma _ (fun y => P)) ..))
  (at level 200, x binder, y binder, right associativity,
  format "'[  ' '[  ' Σ  x  ..  y ']' ,  '/' P ']'") : type_scope.
From EquationsReloaded Require Import Sized nestedrec sized_rec.
Hint Constructors In : core.

Lemma app_in {A} (l l' : list A) x : In x (l ++ l') <-> In x l \/ In x l'.
Proof.
  induction l; simpl; auto.
  split; intros H; depelim H; intuition auto. depelim H.
  split; intros H; depelim H; intuition auto. depelim H; auto.
Qed.
Check Sized.
(* end hide *)
(** ** Well-founded nested recursion as mutual recursion with sizes

  The following example uses just dependent elimination on a finite type (booleans)
  and shows that this also applies to nested recursive definitions. We use
  a simpler type of rose trees here. *)

(** We first define rose trees and their [Sized] instance. *)

Section RoseMut. Context {A : Set}.

  Inductive t : Set :=
  | leaf (a : A) : t
  | node (l : list t) : t.

  Equations t_size : Sized t by struct :=
  t_size (leaf _) := 0;
  t_size (node l) := S (size l).
  (* begin hide *)
  Existing Instance t_size.
  (* end hide *)

  (** The measure just takes the size. Here we encode the mutuality using branching
      on a boolean. *)

  Equations mutmeasure (b : bool) (arg : if b then t else list t) : nat :=
  mutmeasure true t := size t;
  mutmeasure false lt := size lt.

  (** The argument and return type depend on the function label ([true] or [false] here)
      and any well-founded recursive call is allowed. *)

  Equations? elements (b : bool) (x : if b then t else list t) : if b then list A else list A
    by wf (mutmeasure b x) lt :=
  elements true (leaf a) := [a];
  elements true (node l) := elements false l;
  elements false nil := nil;
  elements false (cons t ts) := elements true t ++ elements false ts.
  Proof. all:cbn; lia. Qed.

  (** Dependent return types also possible of course. [elements_dep] is a trivial copy
      of [elements] that additionally shows it is computing a sublist of [elements].
      It requires a dependent return type [elements_dep_type]. *)

  Equations elements_dep_type (b : bool) (x : if b then t else list t) : Set :=
    elements_dep_type true t := { l' : list A | forall x, In x l' -> In x (elements true t) };
    elements_dep_type false l := { l' : list A | forall x, In x l' -> In x (elements false l) }.

  (** We put ourselves in [Program] mode to have coercions of subset types.
      We reset the default obligation tactic which is otherwise applied to
      initial goals when using [Equations?]. *)

  Obligation Tactic := idtac.
  Set Program Mode.

  Equations? elements_dep (b : bool) (x : if b then t else list t) : elements_dep_type b x
    by wf (mutmeasure b x) lt :=
  elements_dep true (leaf a) := [a];
  elements_dep true (node l) := elements_dep false l;
  elements_dep false nil := nil;
  elements_dep false (cons t ts) := elements_dep true t ++ elements_dep false ts.
  Proof.
    all:(simp mutmeasure elements).
    2-3:(cbn; lia). all:destruct elements_dep; simpl. apply i.
    destruct elements_dep;simpl. intros. rewrite app_in in *. intuition.
  Qed.

End RoseMut.
