(** printing <= $\Leftarrow$ *)
(** printing <mul $<_{m}$ *)
(** printing <=mul $\le_{m}$ *)
(** printing =mul= $=_{m}$ *)
(** printing {{ $\{$ *)
(** printing }} $\}$ *)

(* begin hide *)
Set Asymmetric Patterns.
Set Implicit Arguments.
Unset Strict Implicit.

Require Import Arith Utf8.
Require Import Equations.Equations.
(* end hide*)

(*************************************************************************)
(*                          Definition of Fsub                           *)
(*************************************************************************)

(** ** Definition of terms *)

(** Recall that Predicative System F is a typed lambda calculus with
  type abstractions and applications. Our type structure is very simple
  here, with just the function space and universal quantification on
  kinded type variables. We use an absolutely standard de Bruijn
  encoding for type and term variables. The kinds (a.k.a universe
  levels) are represented using natural numbers. Our development is
  based on Jérôme Vouillon's solution to the POPLmark challenge for
  System %$F^{sub}$% %\cite{vouillonpoplmark}%. *)

Definition kind := nat.

Inductive typ : Set :=
  | tvar : nat -> typ | arrow : typ -> typ -> typ | all : kind -> typ -> typ.
(* begin hide *)
Derive NoConfusion for typ.
Instance eqdec_typ : EqDec typ.
Proof. unfold EqDec; repeat decide equality. Defined.
(* end hide *)
(** We will write [∀ X :* k. T] for the quantification over types of
    kind [k]. *)

Inductive term : Set :=
  | var : nat -> term
  | abs : typ -> term -> term
  | app : term -> term -> term
  | tabs : kind -> term -> term
  | tapp : term -> typ -> term.
(* begin hide *)
Derive NoConfusion for term.
Instance eqdec_term : EqDec term.
Proof. unfold EqDec; repeat decide equality. Defined.
(* end hide *)

(** Our raw terms are simply the abstract syntax trees. *)
(** ** Shiftings and substitutions *)

(** We define the different operations of shifting and substitutions
    with the %\Equations% package, we only show the substitution function
    here which uses a [with] right-hand side. All the development can be downloaded
    or browsed at %\url{http://equations-fpred.gforge.inria.fr}%. *)

(* begin hide *)
Equations tshift (X : nat) (T : typ) : typ :=
tshift X (tvar  Y)     => tvar (if le_gt_dec X Y then 1 + Y else Y);
tshift X (arrow T1 T2) => arrow (tshift X T1) (tshift X T2);
tshift X (all   k T2)  => all k (tshift (S X) T2).

Equations shift (x : nat) (t : term) : term :=
shift x (var  y)     => var (if le_gt_dec x y then 1 + y else y);
shift x (abs  T1 t2) => abs T1 (shift (S x) t2);
shift x (app  t1 t2) => app (shift x t1) (shift x t2);
shift x (tabs k  t2) => tabs k (shift x t2);
shift x (tapp t1 T2) => tapp (shift x t1) T2.

Equations shift_typ (X : nat) (t : term) : term :=
shift_typ _ (var y)      => var y;
shift_typ X (abs T1 t2)  => abs (tshift X T1) (shift_typ X t2);
shift_typ X (app t1 t2)  => app (shift_typ X t1) (shift_typ X t2);
shift_typ X (tabs k t2)  => tabs k (shift_typ (1 + X) t2);
shift_typ X (tapp t1 T2) => tapp (shift_typ X t1) (tshift X T2).

Equations tsubst (T : typ) (X : nat) (T' : typ) : typ :=
tsubst (tvar Y) X T' with lt_eq_lt_dec Y X => {
  | inleft (left  _) => tvar Y;
  | inleft (right _) => T';
  | inright _        => tvar (Y - 1) };
tsubst (arrow T1 T2) X T' => arrow (tsubst T1 X T') (tsubst T2 X T');
tsubst (all k T2)    X T' => all k (tsubst T2 (1 + X) (tshift 0 T')).
(* end hide *)

Check shift_typ : ∀ (X : nat) (t : term), term.
Check tsubst : typ -> nat -> typ -> typ.

Equations subst (t : term) (x : nat) (t' : term) : term :=
subst (var y) x t' with lt_eq_lt_dec y x => {
  | inleft (left  _) => var y;
  | inleft (right _) => t';
  | inright _        => var (y - 1) };
subst (abs T1 t2)  x t' => abs T1 (subst t2 (1 + x) (shift 0 t'));
subst (app t1 t2)  x t' => app (subst t1 x t') (subst t2 x t');
subst (tabs k t2)  x t' => tabs k (subst t2 x (shift_typ 0 t'));
subst (tapp t1 T2) x t' => tapp (subst t1 x t') T2.

(* begin hide *)
Equations subst_typ (t : term) (X : nat) (T : typ) : term :=
subst_typ (var y)      _ _ => var y;
subst_typ (abs T1 t2)  X T => abs (tsubst T1 X T) (subst_typ t2 X T);
subst_typ (app e1 e2)  X T => app (subst_typ e1 X T) (subst_typ e2 X T);
subst_typ (tabs k e1)  X T => tabs k (subst_typ e1 (1 + X) (tshift 0 T));
subst_typ (tapp e1 T2) X T => tapp (subst_typ e1 X T) (tsubst T2 X T).
(* end hide *)

Check subst_typ : term -> nat -> typ -> term.

(* begin hide *)

Equations opt_bind {A B : Set} (x : option A) (f : A -> option B) : option B :=
opt_bind (Some x) f => f x;
opt_bind None    _ => None.

Lemma opt_bind_some :
  forall (A B : Set) (e : option A) (f : A -> option B) (b : B),
  opt_bind e f = Some b -> exists a, e = Some a /\ f a = Some b.
Proof.
  intros A B e f b E; funelim (opt_bind e f).
    - exists a; auto.
    - discriminate.
Qed.
(* end hide *)

(** ** Contexts *)

(** We define the contexts [env] and the two functions [get_kind] and
   [get_var] which access the context. A context is an interleaving of
   types and terms contexts. Vouillon's great idea is to have parallel
   de Bruijn indexings for type and term variables, which means separate
   indices for type and term variables. That way, shifting and
   substitution of one kind does not influence the other, making weakening
   and substitution lemmas much simpler, we follow this idea here. *)

Inductive env : Set :=
  | empty : env | evar : env -> typ -> env | etvar : env -> kind -> env.
(* begin hide *)
Instance eqdec_env : EqDec env.
Proof. unfold EqDec; repeat decide equality. Defined.
(* end hide *)

(** Note that %\Equations% allows wildcards and overlapping clauses with a
  first match semantics, as usual. *)

Equations get_kind (e : env) (X : nat) : option kind :=
get_kind  empty       _    => None;
get_kind (evar e _)   X    => get_kind e X;
get_kind (etvar _ T)  O    => Some T;
get_kind (etvar e _) (S X) => get_kind e X.

(** We need the functorial map on [option] types to ease writing these
  partial lookup functions. *)

Equations opt_map {A B : Set} (f : A -> B) (x : option A) : option B :=
opt_map f (Some x) => Some (f x);
opt_map _  None    => None.

Equations get_var (e : env) (x : nat) : option typ :=
get_var  empty       _    => None;
get_var (etvar e _)  x    => opt_map (tshift 0) (get_var e x);
get_var (evar  _ T)  O    => Some T;
get_var (evar  e _) (S x) => get_var e x.

(** ** Well-formedness conditions

  We also define some well-formedness conditions for types, terms
  and contexts. Namely, in a type (resp. in a term), the variables
  must all be kinded (resp. typed). We just show the [wf_typ] definition here,
  those follow Stump and Haye's work. *)

Equations wf_typ (e : env) (T : typ) : Prop :=
wf_typ e (tvar X)      => get_kind e X <> None;
wf_typ e (arrow T1 T2) => wf_typ e T1 /\ wf_typ e T2;
wf_typ e (all k T2)    => wf_typ (etvar e k) T2.

(* begin hide *)
Equations wf_term (e : env) (t : term) : Prop :=
wf_term e (var x)      => get_var e x <> None;
wf_term e (abs T1 t2)  => wf_typ e T1  /\ wf_term (evar e T1) t2;
wf_term e (app t1 t2)  => wf_term e t1 /\ wf_term e t2;
wf_term e (tabs k t)   => wf_term (etvar e k) t;
wf_term e (tapp t1 T2) => wf_term e t1 /\ wf_typ e T2.

Equations wf_env (e : env) : Prop :=
wf_env  empty      => True;
wf_env (evar e T)  => wf_typ e T /\ wf_env e;
wf_env (etvar e k) => wf_env e.
(* end hide *)

(** ** Kinding and typing rules

  The kinding rules are the main difference between Leivant's and
  Stump's presentations. We refer to these works for pen and paper
  presentations of these systems, due to lack of space, we cannot
  include them here. The case for universal quantification sets the
  level of a universal type at [1 + max k k'], where [k] and [k'] are
  respectively the domain and codomain kinds, in Stump's case, which
  allows for a straightforward order on types based on levels, but this
  means that each level is not closed under products from lower levels
  anymore. In other words, multiple quantifications at the same level
  raise the overall level. For example [(∀ X :* 0. X) :* 1] as expected
  but [∀ X :* 0. ∀ Y :* 0. X :* (1 + max 0 (1 + max 0 0)) = 2]. This is
  a very strange behavior.

  We use the standard predicative product rule which sets the product
  level to [max (k+1) k'], which directly corresponds to Martin-Löf's
  Predicative Type Theory. Note that the system includes cumulativity
  through the Var rule which allows to lift a type variable declared at
  level [k] into any higher level [k'].  *)

Inductive kinding : env -> typ -> kind -> Prop :=
| T_TVar : forall (e : env) (X : nat) (k k' : kind), wf_env e -> 
   get_kind e X = Some k -> k <= k' -> kinding e (tvar X) k'
| T_Arrow : forall e T U k k', kinding e T k -> kinding e U k' -> kinding e (arrow T U) (max k k')
| T_All : forall e T k k', kinding (etvar e k) T k' -> kinding e (all k T) (max (k+1) k').
(* begin hide *)
Derive Signature for kinding.
(* end hide *)

(** The typing relation is straightforward. Just note that we check 
  for well-formedness of environments at the variable case, so typing
  derivations are always well-formed. *)

Inductive typing : env -> term -> typ -> Prop :=
  | T_Var (e : env) (x : nat) (T : typ) : wf_env e -> get_var e x = Some T -> typing e (var x) T
  | T_Abs (e : env) (t : term) (T1 T2 : typ) : 
      typing (evar e T1) t T2 -> typing e (abs T1 t) (arrow T1 T2)
  | T_App (e : env) (t1 t2 : term) (T11 T12 : typ) :
      typing e t1 (arrow T11 T12) -> typing e t2 T11 -> typing e (app t1 t2) T12
  | T_Tabs (e : env) (t : term) (k : kind) (T : typ) :
      typing (etvar e k) t T -> typing e (tabs k t) (all k T)
  | T_Tapp (e : env) (t : term) k (T1 T2 : typ) :
      typing e t (all k T1) -> kinding e T2 k -> typing e (tapp t T2) (tsubst T1 0 T2).
(* begin hide *)
Derive Signature for typing.
(* end hide *)
