(* begin hide *)
Set Asymmetric Patterns.
Set Implicit Arguments.
Unset Strict Implicit.

Require Import Arith Omega.
Require Import Equations.Equations.

Require Import Fpred.Tactics.
Require Import Fpred.Definitions.
(* end hide *)

(*************************************************************************)
(*                            General lemmas                             *)
(*************************************************************************)

(*** Commutation properties of shifting and substitution ***)

(* begin hide *)
Ltac tvar_case :=
  match goal with
  | |- ?x =>
      match x with
      | context [le_gt_dec ?n ?n'] =>
          case (le_gt_dec n n')
      | context [(lt_eq_lt_dec ?n ?n')] =>
          case (lt_eq_lt_dec n n'); [intros s; case s; clear s | idtac ]
      end
  end.

Ltac common_cases_simpl :=
  simp tsubst tshift;
  repeat (progress repeat tvar_case; intros; simp tsubst tshift;
          try solve [f_equal; omega]);
  f_equal; trivial.
Ltac common_cases e := funelim e; common_cases_simpl.
(* end hide *)

Lemma tsubst_tshift_prop :
  forall (n : nat) (T T' : typ), T = tsubst (tshift n T) n T'.
Proof.
  intros n T T'; common_cases (tshift n T).
Qed.

Lemma tshift_tshift_prop_1 :
  forall (n n' : nat) (T : typ),
  tshift n (tshift (n + n') T) = tshift (1 + (n + n')) (tshift n T).
Proof.
  intros n n' T; common_cases (tshift n T).
    apply H.
Qed.

Lemma tshift_tsubst_prop_1 :
  forall (n n' : nat) (T T' : typ),
  tshift n (tsubst T (n + n') T') =
  tsubst (tshift n T) (1 + n + n') (tshift n T').
Proof.
  intros n n' T T'; common_cases (tshift n T).
    rewrite (tshift_tshift_prop_1 0 _ T'); apply H.
Qed.

Lemma tshift_tsubst_prop_2 :
  forall (n n' : nat) (T T' : typ),
  (tshift (n + n') (tsubst T n T')) =
  (tsubst (tshift (1 + n + n') T) n (tshift (n + n') T')).
Proof.
  intros n n' T T'; common_cases (tshift (1 + n + n') T).
     rewrite (tshift_tshift_prop_1 0 (n + n')); apply (H (S n)); reflexivity.
Qed.

Lemma tsubst_tsubst_prop :
  forall (n n' : nat) (T U V : typ),
  (tsubst (tsubst T n U) (n + n') V) =
  (tsubst (tsubst T (1 + n + n') (tshift n V)) n (tsubst U (n + n') V)).
Proof.
  intros n n' T U V;
  common_cases (tsubst T (1 + n + n') (tshift n V)).
    - rewrite (tshift_tsubst_prop_1 0 (n + n')).
      rewrite (tshift_tshift_prop_1 0).
      specialize (H (S n) n' t1 (tshift 0 U) (tshift 0 V)).
      rewrite (tshift_tshift_prop_1 0 n) in *. now apply H.
    - apply tsubst_tshift_prop.
Qed.

(****)

(*** Some properties of the well-formedness conditions *)

Lemma wf_typ_env_weaken :
  forall (T : typ) (e e' : env),
  (forall (X : nat), get_kind e' X = None -> get_kind e X = None) ->
  wf_typ e T -> wf_typ e' T.
Proof.
  intros; funelim (wf_typ e T); simp wf_typ in *; eauto.
    - destruct H2; eauto.
    - apply H; auto; intros; destruct X; simp get_kind.
Qed.

Lemma wf_typ_extensionality :
  forall (T : typ) (e e' : env),
  (forall (X : nat), get_kind e X = get_kind e' X) ->
  wf_typ e T -> wf_typ e' T.
Proof.
  intros T e e' H1 H2; apply wf_typ_env_weaken with (2 := H2);
  intros n H3; rewrite H1; trivial.
Qed.

(****)

(*** Removal of a term variable binding ***)

(*
   This corresponds to the environment operation
      E, x : T, E'  |-->  E.
*)

Equations remove_var (e : env) (x : nat) : env :=
remove_var  empty       _    := empty;
remove_var (etvar e' T) x    := etvar (remove_var e' x) T;
remove_var (evar e _)   O    := e;
remove_var (evar e T)  (S x) := evar (remove_var e x) T.

Lemma get_var_remove_var_lt :
  forall (e : env) (x x' : nat),
  x < x' -> get_var (remove_var e x') x = get_var e x.
Proof.
  intros; funelim (remove_var e x'); simp get_var remove_var; auto.
    - inversion H.
    - destruct x; simp get_var; auto; apply H; omega.
    - f_equal; auto.
Qed.

Lemma get_var_remove_var_ge :
  forall (e : env) (x x' : nat),
  x >= x' -> get_var (remove_var e x') x = get_var e (1 + x).
Proof.
  intros; funelim (remove_var e x'); simp get_var remove_var; auto.
    - destruct x; simp get_var; auto; [|apply H]; omega.
    - f_equal; auto.
Qed.

Lemma get_kind_remove_var :
  forall (e : env) (X x': nat), get_kind e X = get_kind (remove_var e x') X.
Proof.
  intros; funelim (remove_var e x'); simp get_var remove_var; auto;
  destruct X; now simp get_kind.
Qed.

Lemma wf_typ_remove_var :
  forall (e : env) (x : nat) (T : typ),
  wf_typ e T -> wf_typ (remove_var e x) T.
Proof.
  intros e n T; apply wf_typ_extensionality;
  intro n'; apply get_kind_remove_var.
Qed.

Lemma wf_typ_insert_var :
  forall (e : env) (n : nat) (T : typ),
  wf_typ (remove_var e n) T -> wf_typ e T.
Proof.
  intros e n T; apply wf_typ_extensionality; intro n'.
  apply sym_eq; apply (get_kind_remove_var e).
Qed.

Lemma wf_env_remove_var :
  forall (e : env) (x : nat), wf_env e -> wf_env (remove_var e x).
Proof.
  intros; funelim (remove_var e x); simp wf_env wf_typ in *; 
  intuition auto using wf_typ_remove_var.
Qed.

(****)

(*** Insertion of a type variable binding in an environment ****)

(*
   This corresponds to the environment operation
       E, E'  |-->  E, X <: T, E'.
   Note that all type variable in E' has to be shifted.
*)

Inductive insert_kind : nat -> env -> env -> Prop :=
    ib_here :
      forall (k : kind) (e : env), insert_kind 0 e (etvar e k)
  | ib_var :
      forall (X : nat) (T : typ) (e e' : env),
      insert_kind X e e' ->
      insert_kind X (evar e T) (evar e' (tshift X T))
  | ib_kind :
      forall (X : nat) (k : kind) (e e' : env),
      insert_kind X e e' ->
      insert_kind (1 + X) (etvar e k) (etvar e' k).
(* begin hide *)
Derive Signature for insert_kind.
(* end hide *)
Derive NoConfusion for env.

Lemma get_kind_insert_kind_ge :
  forall (X X' : nat) (e e' : env),
  insert_kind X' e e' -> X' <= X ->
  get_kind e' (1 + X) = get_kind e X.
Proof.
  intros; funelim (get_kind e X); depelim_ind insert_kind;
  simp get_kind; try omega; auto.
  match goal with H : forall _, _ |- get_kind ?e _ = _ => eapply (H _ e) end;
  eassumption || omega.
Qed.

Lemma get_kind_insert_kind_lt :
  forall (X X' : nat) (e e' : env),
  insert_kind X' e e' -> X < X' ->
  get_kind e' X = get_kind e X.
Proof.
  intros; funelim (get_kind e X); depelim_ind insert_kind;
  simp get_kind; try omega; auto;
  match goal with H : forall _, _ |- get_kind ?e _ = _ => eapply (H _ e) end;
  eassumption || omega.
Qed.

Lemma get_kind_insert_kind :
  forall (X X' : nat) (e e' : env),
  insert_kind X' e e' ->
  get_kind e' (if le_gt_dec X' X then S X else X) = get_kind e X.
Proof.
  intros n n' e e' H;
  destruct (le_gt_dec n' n);
  eauto using get_kind_insert_kind_ge,get_kind_insert_kind_lt.
Qed.

Lemma get_var_insert_kind :
  forall (x X' : nat) (e e' : env),
  insert_kind X' e e' -> get_var e' x = opt_map (tshift X') (get_var e x).
Proof.
  intros; funelim (get_var e' x); depelim_ind insert_kind;
  simp get_var; auto;
  rewrite (H _ _ H0); destruct (get_var e1 x);
  simp opt_map; auto; f_equal; apply (tshift_tshift_prop_1 0 X).
Qed.

Lemma insert_kind_wf_typ :
  forall (T : typ) (X : nat) (e e' : env),
  insert_kind X e e' -> wf_typ e T -> wf_typ e' (tshift X T).
Proof.
  intros; funelim (tshift X T); simp wf_typ in *;
  intuition eauto using insert_kind.
  erewrite get_kind_insert_kind in H1; eauto.
Qed.

Lemma insert_kind_wf_env :
  forall (X : nat) (e e' : env),
  insert_kind X e e' -> wf_env e -> wf_env e'.
Proof.
  intros; funelim (wf_env e'); trivial; depelim H0;
  simp wf_env in *; intuition eauto using insert_kind_wf_typ.
Qed.

(****)

(*** More properties of the well-formedness conditions *)

Lemma wf_typ_weakening_kind :
  forall (e : env) (T : typ) (k : kind),
  wf_typ e T -> wf_typ (etvar e k) (tshift 0 T).
Proof.
  intros e T U H1; apply insert_kind_wf_typ with (2 := H1);
  apply ib_here; trivial.
Qed.

Lemma wf_typ_weakening_var :
  forall (e : env) (T U : typ),
  wf_typ e U -> wf_typ (evar e T) U.
Proof.
  intros e T U H; apply wf_typ_env_weaken with (2 := H); simpl; trivial.
Qed.

Lemma wf_typ_strengthening_var :
  forall (e : env) (T U : typ),
  wf_typ (evar e T) U -> wf_typ e U.
Proof.
  intros e T U H; apply wf_typ_env_weaken with (2 := H); simpl; trivial.
Qed.

Lemma wf_typ_etvar :
  forall (T : typ) (k k' : kind) (e : env) (Hk : k' <= k),
  wf_typ (etvar e k) T -> wf_typ (etvar e k') T.
Proof.
  intros T k k' e Hk; apply wf_typ_env_weaken; intro X; induction X.
    - intros; discriminate.
    - trivial.
Qed.

Lemma get_var_wf :
  forall (e : env) (n : nat) (T : typ),
  wf_env e -> get_var e n = Some T -> wf_typ e T.
Proof.
  intros; funelim (get_var e n); simp wf_env get_var in *;
  repeat match goal with
    | H : context [ opt_map _ (get_var ?e ?x) ] |- _ =>
        destruct (get_var e x); simp opt_map in H
    | H : Some _ = Some _ |- _ => depelim H
    end; try discriminate;
    intuition auto using wf_typ_weakening_var, wf_typ_weakening_kind.
Qed.