(** printing < $<$ *)
(* begin hide *)
Set Asymmetric Patterns.
Set Implicit Arguments.
Unset Strict Implicit.

Require Import Arith Omega.
Require Import Equations.Equations.

Require Import Fpred.Tactics.
Require Import Fpred.Definitions.
Require Import Fpred.General.
(* end hide *)
(** The metatheory of the system is pretty straightforward and follows
  the one of %$F^{sub}$% closely. We only mention the main idea for type 
  substitution and the statements of the main metatheoretical lemmas. *)

(** To formalize type substitution, we use a proposition [env_subst]
    that corresponds to the environment operation: [E, X :* k, E']
    %$\Rightarrow$% [E, (X] %$\mapsto$% [T') E'] assuming [E |- T' :*
    k].  In other words, [env_subst X T e e'] holds whenever we can find
    environments [E], [E'] and a kind [k] such that [E |- T' :* k] and
    [e = E, X :* k, E'] and [e' = E, (X] %$\mapsto$% [T') E'].  *)

Inductive env_subst : nat -> typ -> env -> env -> Prop :=
| es_here (e : env) (T : typ) (k : kind) : kinding e T k -> env_subst 0 T (etvar e k) e
| es_var (X : nat) (T T' : typ) (e e' : env) : 
    env_subst X T' e e' -> env_subst X T' (evar e T) (evar e' (tsubst T X T'))
| es_kind (X : nat) k (T' : typ) (e e' : env) :
    env_subst X T' e e' -> env_subst (1 + X) (tshift 0 T') (etvar e k) (etvar e' k).
(* begin hide *)
Derive Signature for env_subst.
(* end hide *)

(* begin hide *)
Lemma env_subst_get_var :
  forall (x X' : nat) (e e' : env) (T : typ),
  (env_subst  X' T e e') ->
  get_var e' x = opt_map (fun T' => tsubst T' X' T) (get_var e x).
Proof.
  intros; funelim (get_var e' x); depelim_ind env_subst;
  simp get_var opt_map;
  repeat match goal with
  | H : _ |- _ => erewrite H; [ | eassumption]
  | |- context[ get_var ?e ?x ] => destruct (get_var e x)
  end; simp opt_map; f_equal;
  try solve [apply tsubst_tshift_prop];
  apply (tshift_tsubst_prop_1 0 X).
Qed.

Lemma env_subst_get_kind_lt :
  forall (X X' : nat) (e e' : env) (T : typ),
  (env_subst X' T e e') -> X < X' ->
  get_kind e' X = get_kind e X.
Proof.
  intros n n' e e' T H; funelim (get_kind e' n); depelim_ind env_subst; auto;
  intros; simp get_kind; try omega;
  match goal with H : _ |- _ => eapply H; eauto with arith end.
Qed.

Require Import Program.Tactics.
Ltac simplify_goal := repeat revert_last; repeat (simplify ? || intro).

Lemma env_subst_get_kind_ge :
  forall (X X' : nat) (e e' : env) (T : typ),
  env_subst X' T e e' -> X' < X ->
  get_kind e' (X - 1) = get_kind e X.
Proof.
  intros n n' e e' T H; funelim (get_kind e' (n - 1)); depelim_ind env_subst;
  intros; simp get_kind;
  try match goal with H : _ < ?n |- _ => destruct n; try omega end;
  simp get_kind; simpl in *; try rewrite <- minus_n_O in *; simplify_goal; trivial.
  eapply H; eauto. f_equal; omega. omega.
Qed.
(* end hide *)

(** ** Typing and well-formedness *)

(** Actually, both kinding and typing imply well-formedness. In other
   words, it is possible to kind a type [T] in an environment [e] only
   if both the type and the environment are well-formed. *)

Lemma kinding_wf (e : env) (T : typ) (k : kind) : kinding e T k -> wf_env e /\ wf_typ e T.
Proof.
  intros; funelim (wf_typ e T); simplify_goal; repeat split; depelim_ind kinding;
  try match goal with H : _ |- _ => solve [eapply H; eassumption] end;
  match goal with H : _ = _ |- _ => rewrite H end; discriminate.
Qed.

(* begin hide *)
Definition kinding_wf_env e t U (H : kinding e t U) : wf_env e := proj1 (kinding_wf H).
Definition kinding_wf_typ e t U (H : kinding e t U) : wf_typ e t := proj2 (kinding_wf H).
Hint Resolve kinding_wf_env kinding_wf_typ.

Lemma env_subst_wf_typ_0 :
  forall (e e' : env) (T : typ) (X : nat),
  env_subst X T e e' -> wf_env e' -> wf_typ e' T.
Proof.
  intros e e' T X H; induction H; simpl.
    - intros _; exact (proj2 (kinding_wf H)).
    - intros (H1, H2); apply wf_typ_weakening_var; auto.
    - intros H2; apply wf_typ_weakening_kind; auto.
Qed.

Lemma env_subst_wf_typ :
  forall (e e' : env) (S T : typ) (X : nat),
  env_subst X T e e' -> wf_typ e S -> wf_env e' -> wf_typ e' (tsubst S X T).
Proof.
  intros; funelim (tsubst S X T); simp wf_typ in *; repeat split;
  try match goal with
  | H : _ |- _ => solve [eapply H; intuition eauto using env_subst]
  end.
    - rewrite (env_subst_get_kind_lt H); trivial.
    - eapply env_subst_wf_typ_0; intuition.
    - rewrite (env_subst_get_kind_ge H); trivial.
Qed.

(** Using the [env_subst] judgment, we can easily show that substitution
  in an environment preserves the well-formedness, kinding and typing judgments. *)

Lemma env_subst_wf_env (e e' : env) (T : typ) (X : nat) : env_subst X T e e' -> wf_env e -> wf_env e'.
Proof.
  intros H; induction H; simpl; try tauto.
  simp wf_env.
  intros (H1, H2). split; auto.
  apply env_subst_wf_typ with (1 := H); auto.
Qed.

(*** Typing relation well-formedness ***)

(** Same as kinding, it is possible to type a term in an environment
    only if the term, its type and the environment are well-formed. *)

Lemma typing_wf :
  forall (e : env) (t : term) (U : typ),
  typing e t U -> wf_env e /\ wf_term e t /\ wf_typ e U.
Proof.
  intros; funelim (wf_term e t); depelim_ind typing;
  simp wf_term wf_typ wf_env in *;
  repeat match goal with
  | H : forall _, typing ?e ?t _ -> _,
    H' : typing ?e ?t ?T |- _ => specialize (H T)
  end; simp wf_typ in *;
  intuition eauto using get_var_wf, wf_typ_strengthening_var,
  env_subst_wf_typ, env_subst;
  match goal with H1 : _ = _, H2 : _ = _  |- _ => rewrite H1 in H2 end;
  discriminate.
Qed.

Definition typing_wf_env e t U (H : typing e t U) : wf_env e := proj1 (typing_wf H).
Definition typing_wf_typ e t U (H : typing e t U) : wf_typ e U := proj2 (proj2 (typing_wf H)).
Hint Resolve typing_wf_env typing_wf_typ.
(* end hide *)

(** ** Weakening *)

(* begin hide *)
Lemma kinding_weakening_kind_ind :
  forall (e e' : env) (X : nat) (U : typ) k,
  insert_kind X e e' -> kinding e U k -> kinding e' (tshift X U) k.
Proof.
  intros e e' X U k H1 H2; funelim (tshift X U); depelim_ind kinding;
  econstructor; eauto using kinding, insert_kind, insert_kind_wf_env.
  erewrite get_kind_insert_kind; eauto.
Qed.

(** Weakening by a type variable preserves kinding. *)

Lemma kinding_weakening_kind (e : env) (T : typ) k k' : kinding e T k -> kinding (etvar e k') (tshift 0 T) k.
Proof.
  intros H; apply kinding_weakening_kind_ind with (2 := H);
  apply ib_here; trivial.
Qed.

(*
   Rather than proving the weakening property for term variable
   bindings by induction on a derivation on a derivation, as in the
   paper proof, we prove the more general result that the subtyping
   relation does not depend on the term variable bindings.  Then, this
   intermediate result can be used for proving strengthening.
*)

(** We will need an extensionality lemma on kinding in a context to prove
    the next weakening result. *)

Lemma kinding_extensionality :
  forall (e e' : env) (U : typ) k,
  (forall (X : nat), (get_kind e X) = (get_kind e' X)) -> wf_env e' ->
  kinding e U k -> kinding e' U k.
Proof.
  intros e e' Y k H1 H2 H3; revert e' H1 H2;
  induction H3; intros; econstructor; eauto.
    - rewrite <- H2; trivial.
    - apply IHkinding; trivial; destruct X; now simp get_kind.
Qed.

Lemma kinding_weakening_var_ind :
  forall (e : env) (x : nat) (U : typ) k,
  wf_env e -> kinding (remove_var e x) U k -> kinding e U k.
Proof.
  intros e x U V H1; apply kinding_extensionality; trivial;
  intro X; apply sym_eq; apply (get_kind_remove_var e).
Qed.

(* Weakening by a term variable preserves kinding. *)

Lemma kinding_weakening_var :
  forall (e : env) (T V : typ) k,
  wf_typ e V -> kinding e T k -> kinding (evar e V) T k.
Proof.
  intros e T V k H1 H2; generalize H2; apply kinding_extensionality; trivial;
  simpl; split; trivial; exact (proj1 (kinding_wf H2)).
Qed.

(* end hide *)

(** We only show the main weakening lemma for typing: if [e'] results from
  [e] by inserting a type variable at position [X] with any kind,
  the term and types of a typing derivation can be shifted accordingly
  to give a new typing derivation in the extended environment. *)

Lemma typing_weakening_kind_ind (e e' : env) (X : nat) (t : term) (U : typ) :
  insert_kind X e e' -> typing e t U -> typing e' (shift_typ X t) (tshift X U).
Proof.
  intros; funelim (shift_typ X t); depelim_ind typing;
  simp tshift; eauto using typing, insert_kind.
    - constructor; eauto using insert_kind_wf_env;
      rewrite (get_var_insert_kind _ H);
      match goal with H : _ = _ |- _ => rewrite H end; trivial.
    - econstructor; eauto; exact (H _ _ _ H1 H2_).
    - rewrite (tshift_tsubst_prop_2 0 X);
      econstructor; eauto using kinding_weakening_kind_ind;
      exact (H _ _ _ H0 H1).
Qed.

(* begin hide *)

(** More interestingly, weakening by a type variable preserves typing. *)

Lemma typing_weakening_kind :
  forall (e : env) (t : term) (U : typ) k,
  typing e t U ->
  typing (etvar e k) (shift_typ 0 t) (tshift 0 U).
Proof.
  intros e t U V H2; apply typing_weakening_kind_ind with (2 := H2);
  apply ib_here; trivial.
Qed.

Lemma typing_weakening_var_ind :
  forall (e : env) (x : nat) (t : term) (U : typ),
  wf_env e -> typing (remove_var e x) t U -> typing e (shift x t) U.
Proof.
  intros e n t U H1 H2; funelim (shift n t); depelim_ind typing;
  econstructor; eauto using kinding_weakening_var_ind.
    - on_call le_gt_dec ltac:(fun c => destruct c);
      [ match goal with H : _ = _ |- _ => rewrite get_var_remove_var_ge in H end
      | match goal with H : _ = _ |- _ => rewrite get_var_remove_var_lt in H end ];
      trivial.
    - apply H; eauto; split; trivial;
      eapply wf_typ_insert_var; apply (typing_wf H2).
Qed.
(* end hide *)

(** Weakening by a term variable preserves typing as well. *)

Lemma typing_weakening_var (e : env) (t : term) (U V : typ) :
  wf_typ e V -> typing e t U -> typing (evar e V) (shift 0 t) U.
Proof.
  intros H1 H2; apply (@typing_weakening_var_ind (evar e V));
  simpl; trivial; split; trivial; exact (proj1 (typing_wf H2)).
Qed.

(* begin hide *)

(** Removing a term variable binding from a context obviously does not
    alter kinding. *)

Lemma kinding_strengthening_var :
  forall (e : env) (x : nat) (U : typ) k,
   kinding e U k -> kinding (remove_var e x) U k.
Proof.
  intros e x U V H1; generalize H1; apply kinding_extensionality.
    - intro X; apply get_kind_remove_var.
    - apply wf_env_remove_var; exact (proj1 (kinding_wf H1)).
Qed.

(****)

(*** A.7 Lemma [Narrowing for the Typing Relation] ***)
(* end hide *)

(** ** Narrowing *)

(** As the system includes a kind of subtyping relation due to level
  cumulativity, we can prove a narrowing property for derivations. Again
  we define a judgment formalizing that a context [e'] is a narrowing of
  a context [e] if they are identical but for one type variable
  binding [(T : k')] in [e'] and [(T : k)] in [e] with [k' < k]. *)

Inductive narrow : nat -> env -> env -> Set :=
    narrow_0 (e : env) (k k' : kind) : k' < k -> narrow 0 (etvar e k) (etvar e k')
  | narrow_extend_kind (e e' : env) (k : kind) (X : nat) :
      narrow X e e' -> narrow (1 + X) (etvar e k) (etvar e' k)
  | narrow_extend_var (e e' : env) (T : typ) (X : nat) :
      wf_typ e' T -> narrow X e e' -> narrow X (evar e T) (evar e' T).
(* begin hide *)
Derive Signature for narrow.
(* end hide *)

(* begin hide *)
Lemma get_kind_narrow_ne :
  forall (X X' : nat) (e e' : env),
  narrow X e e' -> X' <> X -> get_kind e X' = get_kind e' X'.
Proof.
  intros; funelim (get_kind e X'); depelim_ind narrow;
  simp get_kind; eauto; omega.
Qed.

Lemma get_kind_narrow_eq :
  forall (X : nat) (e e' : env),
  narrow X e e' ->
  { k : kind & { k' |
  get_kind e X = Some k /\ get_kind e' X = Some k' /\ k' < k } }.
Proof.
  intros; induction H.
    - exists k; exists k'; repeat split; trivial.
    - destruct IHnarrow as (Q, (P, (H1, (H2, H3))));
      exists Q; exists P; repeat split; simp get_kind.
    - destruct IHnarrow as (Q, (P, (H1, (H2, H3))));
      exists Q; exists P; repeat split; simp get_kind.
Qed.

Lemma get_var_narrow :
  forall (X x' : nat) (e e' : env),
  narrow X e e' -> get_var e x' = get_var e' x'.
Proof.
  intros; funelim (get_var e x'); depelim_ind narrow;
  simp get_var; auto; rewrite (H _ _ H0); trivial.
Qed.

Lemma narrow_wf_typ :
  forall (e e' : env) (T : typ) (X : nat),
  narrow X e e' -> wf_typ e T -> wf_typ e' T.
Proof.
  intros e e' T n H1 H2;
  apply wf_typ_env_weaken with (2 := H2); intros n' H3;
  case (eq_nat_dec n' n).
    - intros E; rewrite E in H3;
      generalize (get_kind_narrow_eq H1);
      intros (Q, (P, (H4, (H5, H6)))); rewrite H5 in H3; discriminate.
    - intro H4; rewrite (get_kind_narrow_ne H1 H4); trivial.
Qed.

Lemma narrow_wf_env :
  forall (e e' : env) (X : nat),
  narrow X e e' -> wf_env e -> wf_env e'.
Proof.
  intros e e' n H; induction H; simpl;
  intros H1; auto; split; auto; intuition.
Qed.


Definition narrowing_prop (k : kind) :=
  forall (e e' : env) (X : nat) (S : typ) k',
  narrow X e e' -> get_kind e X = Some k ->
  kinding e S k' -> kinding e' S k'.
(* end hide *)

(** Before we can show narrowing, we have to show that kinding respects cumulativity:
   If it is provable that a type [T] has kind [k] in the context [e], then
   we can also prove that it has any kind [k'] for [k <= k']. *)

Lemma kinding_transitive e T k k' : kinding e T k -> k <= k' -> kinding e T k'.
Proof.
  intros H; revert k'; induction H; intros k'' H2.
    - econstructor; eauto with arith.
    - replace k'' with (max k k'') by eauto with arith;
      eauto using kinding with arith.
    - destruct (le_lt_eq_dec _ _ H2).
        + rewrite Nat.max_lub_lt_iff in l;
          replace k'' with (max (k + 1) ((pred k'') + 1)) by
          (replace (pred k'' + 1) with k'' by omega; eauto with arith);
          constructor; eapply IHkinding; omega.
        + subst; eauto using kinding.
Qed.

(* begin hide *)
Lemma narrowing_case :
  forall Q : kind, narrowing_prop Q.
Proof.
  intros Q e e' n T1 T2 H3 H4 H5; generalize dependent n; revert e' Q;
  induction H5; intros; eauto using kinding, narrow_extend_kind;
  destruct (eq_nat_dec X n); subst.
    - destruct (get_kind_narrow_eq H3) as [? [? [HH [? ?]]]];
      rewrite H4 in *. depelim H0; depelim HH.
      econstructor; eauto using narrow_wf_env; omega.
    - econstructor; eauto using narrow_wf_env;
      rewrite <- (get_kind_narrow_ne H3); trivial.
Qed.
(* end hide *)

(** Narrowing is a strong property, in the sense that a type [T] can
    have in a narrowing of a context [e] any kind that it can have in
    [e] itself. *)
(* begin hide *)

Lemma kinding_narrowing :
  forall (e e' : env) (X : nat) (S : typ) k,
  narrow X e e' -> kinding e S k -> kinding e' S k.
Proof.
  intros e e' n S T H1 H2;
  generalize (get_kind_narrow_eq H1); intros (Q, (P, (H3, _)));
  exact (narrowing_case H1 H3 H2).
Qed.
(* end hide *)

Lemma typing_narrowing_ind (e e' : env) (X : nat) (t : term) (U : typ) : narrow X e e' -> typing e t U -> typing e' t U.
Proof.
  intros H1 H2; generalize e' X H1; clear e' X H1;
  induction H2 using typing_ind.
    - intros e' n1 H1; apply T_Var.
        + exact (narrow_wf_env H1 H).
        + rewrite <- get_var_narrow with (1 := H1); trivial.
    - intros e' n1 H1; apply T_Abs; apply IHtyping with (X := n1);
      apply narrow_extend_var; trivial;
      exact (narrow_wf_typ H1 (proj1 (proj1 (typing_wf H2)))).
    - intros e' n1 H1; eapply T_App; eauto.
    - intros e' n1 H1; apply T_Tabs; apply IHtyping with (X := S n1);
      apply narrow_extend_kind with (X := n1); trivial;
      exact (narrow_wf_typ H1 (proj1 (proj1 (typing_wf H2)))).
    - intros e' n1 H1; eapply T_Tapp; eauto; exact (kinding_narrowing H1 H). 
Qed.

(* begin hide *)

(** This allows us to prove that the typing relation is preserved by
    narrowing. *)
    
Lemma typing_narrowing :
  forall (e : env) (t : term) (U : typ) k k',
  typing (etvar e k) t U -> k' < k -> typing (etvar e k') t U.
Proof.
  intros e t U V1 V2 H1 H2; exact (typing_narrowing_ind (narrow_0 _ H2) H1).
Qed.

(****)

(*** A.8 Lemma [Subtyping preserves typing] ***)

(* Compared to the lemma in the paper proof, this lemma is slightly
   stronger: [u] is typed in a larger environment.  This makes it
   possible to use our one-step weakening lemmas rather that the
   stronger lemmas of the paper proof. *)

(* end hide *)

(** ** Substitution *)

(** Now, substitution lemmas can be proven for the various
    substitution functions. *)

Lemma subst_preserves_typing (e : env) (x : nat) (t u : term) (V W : typ) :
  typing e t V -> typing (remove_var e x) u W -> get_var e x = Some W -> typing (remove_var e x) (subst t x u) V.
Proof.
  intros; funelim (subst t x u); (depelim_ind typing; []).
    - constructor. eapply (H _ _ _ H0); eauto.
      apply typing_wf_env in H0; simp wf_env remove_var in *.
      intuition eauto using typing_weakening_var, wf_typ_remove_var.
    - eauto using typing.
    - constructor. eapply (H _ _ _ H0); eauto.
      simp remove_var; apply typing_weakening_kind; eauto.
      simp get_var; rewrite H2; now simp opt_map.
    - eauto using typing, kinding_strengthening_var.
    - constructor; eauto. rewrite get_var_remove_var_lt; trivial.
    - subst x. match goal with
      H1 : _ = _, H2 : _ = _ |- _ => rewrite H1 in H2; depelim H2
      end; trivial.
    - constructor; eauto; rewrite get_var_remove_var_ge; try omega;
      match goal with
        |- context [1 + (?x - 1)] => replace (1 + (x - 1)) with x by omega
      end; trivial.
Qed.

(* begin hide *)
(****)

(*** A.10 Lemma [Type substitution preserves subtyping] ***)

Lemma env_subst_kinding :
  forall (e e' : env) (T : typ) (X : nat) k,
  env_subst X T e e' -> (get_kind e X) = (Some k) -> wf_env e' ->
  kinding e' T k.
Proof.
  intros e e' T X k H; revert k; induction H; intros;
  simp get_kind wf_env in *.
    - depelim H0; trivial.
    - intuition eauto using kinding_weakening_var.
    - auto using kinding_weakening_kind.
Qed.

Lemma tsubst_preserves_kinding :
  forall (e e' : env) (X : nat) (T U : typ) k,
  env_subst X T e e' -> kinding e U k -> kinding e' (tsubst U X T) k.
Proof.
  intros; funelim (tsubst U X T); simplify_goal; depelim_ind kinding;
  eauto using kinding, env_subst, kinding_transitive,
  env_subst_kinding, env_subst_wf_env.
    - econstructor; eauto using env_subst_wf_env.
      rewrite (env_subst_get_kind_lt H); trivial.
    - econstructor; eauto using env_subst_wf_env.
      rewrite (env_subst_get_kind_ge H); trivial.
Qed.

(****)

(*** A.11 Lemma [Type substitution preserves typing] ***)

Lemma subst_typ_preserves_typing_ind :
  forall (e e' : env) (t : term) (U P : typ) (X : nat),
  env_subst X P e e' ->
  typing e t U -> typing e' (subst_typ t X P) (tsubst U X P).
Proof.
  intros; funelim (subst_typ t X P); simplify_goal; depelim_ind typing; simp tsubst.
    - constructor; eauto using env_subst_wf_env.
      erewrite env_subst_get_var; eauto.
      match goal with H : _ = _ |- _ => rewrite H; trivial end.
    - eauto using typing, env_subst.
    - econstructor; eauto. apply (H _ _ (arrow T11 U) H1); trivial.
    - eauto using typing, env_subst.
    - rewrite (tsubst_tsubst_prop 0 X).
      econstructor; eauto using tsubst_preserves_kinding.
      exact (H _ _ _ H0 H1).
Qed.
(* end hide *)

Lemma subst_typ_preserves_typing (e : env) (t : term) (U P : typ) k :
  typing (etvar e k) t U -> kinding e P k -> typing e (subst_typ t 0 P) (tsubst U 0 P).
Proof.
  intros H1 H2;
  exact (subst_typ_preserves_typing_ind (es_here H2) H1).
Qed.

(** Finally, we prove regularity, which is to say that the type of any
    well-typed term is kinded. This is a consequence of the fact that
    any well-formed type is kindable. All these results correspond
    directly to the paper proofs of Stump and Hayes. *)

(* begin hide *)
Lemma wf_typ_kinding :
  forall (e : env) (U : typ), wf_env e -> wf_typ e U -> exists k : kind, kinding e U k.
Proof.
  intros e U H. revert e H. induction U; intros e He.
  simpl; intros H. case_eq (get_kind e n).
    - intros k Hk. exists k. apply T_TVar with k; auto.
    - contradiction.
    - intros [H1 H2]. destruct (IHU1 _ He H1) as [k Hk], (IHU2 _ He H2) as [k' Hk'].
      exists (max k k'). apply T_Arrow; auto.
    - simpl; intros Ht. destruct (IHU (etvar e k) He Ht) as [k' Hk'].
      exists (max (k+1) k'). apply T_All; auto.
Qed.
(* end hide *)

Theorem regularity (e : env) (t : term) (U : typ) : typing e t U -> exists k : kind, kinding e U k.
Proof.
  intros.
  destruct (typing_wf H) as [? [? ?]].
  apply wf_typ_kinding; assumption.
Qed.
