(** printing eqns %\Equations% *)
(* begin hide *)
Require Import Equations.Equations.
Require Import Relations FunctionalExtensionality.
Scheme Acc_dep := Induction for Acc Sort Prop.
Arguments WellFounded A R : clear implicits.
Check Init.Wf.Fix.

Section Wf.
(* end hide *)

(** The definition of [FixWf] is simply a wrapper around the standard library's [Wf.Fix] 
    combinator. *)
(* begin hide *)
  (* We put ourselves in a section assuming a well-founded relation [R] on a type [A]
    and a functional [step] that can perform recursive calls on smaller arguments. *)

Context {A R} {WF : WellFounded A R} (P : A -> Type).
Context (step : forall x : A, (forall y : A, R y x -> P y) -> P x).
  
Definition FixWf : forall x : A, P x := Fix wellfounded P step.

(* We want to prove the general lemma [FixWf_unfold] below, which shows
    that the [step] function's result is actually independent of the way
    its functional argument is implemented. *)

(* end hide *)
(** Previously, the unfolding lemma was proved using functional extensionality
    and proof irrelevance for the accessibility proofs. We can avoid the
    use of proof-irrelevance by showing that accessibility is
    propositionally proof-irrelevant by definition, a folklore result: *)
  
  Lemma Acc_pi (x : A) (p q : Acc R x) : p = q.
  (* begin hide *)
  Proof.
    induction p using Acc_dep in q |- *.
    destruct q.
    f_equal.
    extensionality y. extensionality H'. apply H.
  Qed.
  (* end hide *)

  (** The proof of this lemma is by (dependent) induction on the first
    accessibility proof and destruction of the second one, and requires
    two uses of the axiom of (dependent) functional extensionality, as
    the constructor arguments of [Acc] is a function.

    Hence, our proofs of unfolding for well-founded recursive definitions
    must rely on the functional extensionality axiom. This is the only
    axiom used by %\Equations% now. Note that in a system with
    _definitional_ proof irrelevance, this axiom would disappear as
    well. Finally, the presence of this axiom is not problematic in
    practice, as anyway the unfolding behavior of [FixWf] on open terms,
    even assuming a closed well-foundedness proof, makes it difficult to
    handle during proofs and users tend to resort to unfolding lemmas
    instead. Note also that the unfolding lemma and the elimination
    principle allow reasoning on a recursive function even if its
    termination proof has not been provided, i.e. if it was admitted as
    an axiom. *)

  (* begin hide *)
  
  Lemma FixWf_unfold (x : A) :
    FixWf x = step x (fun y _ => FixWf y).
  Proof.
    intros. unfold FixWf, Fix. destruct wellfounded.
    simpl. f_equal. extensionality y. extensionality h.
    f_equal. apply Acc_pi.
  Qed.
End Wf.
(* end hide *)
