(** printing by %\coqdockw{by}% *)
(** printing rec %\coqdockw{rec}% *)
(* begin hide *)
From Equations Require Import Equations.
From EquationsReloaded Require Import Fin.
Require Import Omega Utf8 Arith Compare_dec.
Set Equations Transparent.
(* end hide *)

(** %\Equations% supports mutual recursion on mutual inductive types at
    the top-level, using a syntax close to vanila %\Coq%. We demonstrate
    it on an example formalization of a term-language for λ-calculus
    with non-empty application spines and well-scoped variables.  We
    define this datatype as a mutual indexed inductive type of [term]s
    and [spine]s. Technically, we use a non-recursively uniform
    parameter here instead of a proper index for the number of free
    variables. That is, the parameter varies in the arguments of the
    constructor, e.g. for [Lam] where we introduce a fresh free variable,
    but not in their conclusions, which we even omit here for [term].
    They are treated the same by %\Equations%. *)

Inductive term (n : nat) : Set :=
| Var (f : fin n) | Lam (t : term (S n)) | App (t : term n) (l : spine n)
with spine (n : nat) : Set :=
| tip : term n -> spine n | snoc : spine n -> term n -> spine n.

(* begin hide *)
Arguments Var {n} f.
Arguments Lam {n} t.
Arguments App {n} t l.
Arguments tip {n} t.
Arguments snoc {n} ts t.
Derive Signature for term spine.
(* end hide *)

(** Suppose we want to define capture-avoiding substitution for this
    language. We first need to define lifting of a well-scoped term with
    [n] variables into a well-scoped term with [n+1] free variables,
    shifting variables above or equal to [k] by [1].
    We handle separately the case of variables: *)

Equations lift_fin {n} (k : nat) (f : fin n) : fin (S n) :=
  lift_fin 0 f := fs f;
  lift_fin (S k) fz := fz;
  lift_fin (S k) (fs f) := fs (lift_fin k f).

(** Lifting a [fin] changes the index to make space for the new free variable.
    Using a toplevel [where] clause, one can define the
    mutually recursive lifting function on terms and spines. The
    definition is accepted by the guard cheker easily. *)

Equations lift {n} (k : nat) (u : term n) : term (S n) := {
  lift k (Var f) => Var (lift_fin k f); lift k (Lam t) => Lam (lift (S k) t);
  lift k (App f ts) => App (lift k f) (lift_spine k ts) }
where lift_spine {n : nat} (k : nat) (t : spine n) : spine (S n) := {
  lift_spine k (tip t) => tip (lift k t);
  lift_spine k (snoc ts t) => snoc (lift_spine k ts) (lift k t) }.

(** This representation lends itself naturally to a definition of
    parallel substitution. To define it at the case of abstraction, one
    must explain how to lift a substitution of [k] variables to a
    substitution of [k+1] variables by preserving the [0th] variable and
    lifting the result of the substitution otherwise.  This
    pattern-matching involves the no-confusion principle on the natural
    number index (constructors are distinct and injective),
    but does not require UIP. *)

Equations extend_var (k : nat) (u : fin (S k) -> term k) (f : fin (S (S k))) : term (S k) :=
  extend_var k u fz => Var fz ;
  extend_var k u (fs f) => lift 0 (u f).

(** Again, using a toplevel [where] clause, we can define this parallel substitution,
    structurally on the term and spine. *)

Equations subst_term {k : nat} (u : fin (S k) -> term k) (t : term (S k)) : term k := {
  subst_term u (Var v) => u v; subst_term u (Lam t) => Lam (subst_term (extend_var k u) t);
  subst_term u (App t l) => App (subst_term u t) (subst_spine u l) }
where subst_spine {k : nat} (u : fin (S k) -> term k) (t : spine (S k)) : spine k := {
  subst_spine u (tip t) => tip (subst_term u t);
  subst_spine u (snoc ts t) => snoc (subst_spine u ts) (subst_term u t) }.

